#!/bin/sh
#
# Configure the library

# Check for debugging build or normal build
if [ "$1" = "-d" ]; then
	debug="on";
	shift;
fi
if [ "$SDL_DEBUG" != "" ]; then
	debug="on"
fi

# The current version of the SDL library.
if [ -f .version ]; then
	version=`cat .version`
else
	echo "No version file!  Is your archive complete?"
	exit 1
fi

# Create a file descriptor for talking to the user.
# (/dev/tty not supported on BeOS PR2)
exec 3>&1

# Echo a line without a trailing newline:
prompt()
{
	line="`echo '1\n2'`"
	if [ "$line" = '1\n2' ]; then
		echo -n "$*" >&3
	else
		echo "$*\c" >&3
	fi
}
# Find a program in the execution path:
findexe()
{
	exe=$1
	ifs=$IFS
	IFS=:
	set -- $PATH
	IFS=$ifs
	while [ $# -gt 0 ];
	do path=$1/$exe
           if [ -f $path ]; then
		echo $path; return
	   fi
	   shift
	done
}
# Strip the directory from a pathname
stripdir()
{
	ifs=$IFS
	IFS=/
	set -- $1
	IFS=$ifs
	while [ $# -gt 1 ]; do shift; done
	echo $1
}

# Let the user select which subsystem drivers to use for SDL
SelectSubsystems()
{
	echo ""
	choice="$*"
	while [ "$choice" != "" ]; do
		if [ "$choice" = "?" ]; then
			echo "" >&3
			for dir in `echo $subsystems | tr '[A-Z]' '[a-z]'`
			do  if [ -r "src/$target/$dir/.help" ]; then
				echo $dir:			>&3
				cat src/$target/$dir/.help	>&3
				echo ""				>&3
			    else
				echo "No help available for $dir on $target" >&3
			    fi
			done
		else
			subsystems=$choice
		fi
		echo "Choose $target interface methods (or ? for help)" >&3
		prompt "[$subsystems] "
		read choice
	done
	echo $subsystems | tr '[A-Z]' '[a-z]'
}

CheckGLIBC()
{
	cat >glibc.c <<__EOF__
#include <stdlib.h>
#include <features.h>
#ifdef __GLIBC__
int main () { exit(0); }
#else
int main () { exit(1); }
#endif
__EOF__
	make glibc >/dev/null 2>&1
	if [ -x glibc ] && ./glibc >/dev/null; then
		status=0
	else
		status=1
	fi
	rm -f glibc.c glibc
	return $status
}
CheckWinGCC()
{
	cat >wincheck.c <<__EOF__
#include <windows.h>
int main () { exit(0); }
__EOF__
	make wincheck >/dev/null 2>&1
	status=$?
	rm -f wincheck.c wincheck wincheck.exe
	return $status
}
CheckDirectX()
{
	cat >dxcheck.c <<__EOF__
#include <directx.h>
int main () { exit(0); }
__EOF__
	make dxcheck CFLAGS="-I./directx/include" >/dev/null 2>&1
	status=$?
	rm -f dxcheck.c dxcheck dxcheck.exe
	return $status
}
CheckXF86DGA()
{
	cat >xf86dga.c <<__EOF__
#include <X11/Xlib.h>
#include <X11/extensions/xf86dga.h>
#include <X11/extensions/xf86vmode.h>
int main () { exit(0); }
__EOF__
	make xf86dga >/dev/null 2>&1
	status=$?
	rm -f xf86dga.c xf86dga
	return $status
}
CheckNASM()
{
	if [ "$nasm" != ":" ]; then
		echo \
"
* Detected iX86 processor and NASM compiler, adding MMX support for $1 "
#
		status=0
	else
		echo \
"
* Couldn't find NASM compiler -- not compiling MMX support for $1
  (NASM is available from: http://www.cryogen.com/Nasm/ ) "
#
		status=1
	fi
	return $status
}
# Generate the flags necessary to create the Makefile
#
# The following global variables must be set:
#
#	$version	The SDL library version
#	$target		The OS target (lowercase)
#	$1		The interface subsystem directory
#
# The following global variables are set by this function:
#
#	$extrainc	Header directories for building shared library
#	$extraflags	Compiler flags for building shared library
#	$sharedlib	The name of the shared library object file
#	$sharedrule	Makefile rule used to build shared library
#
GenMakeflags()
{
	subsystem=$1
	case $target in
	  linux)
		# Allow GCC specific optimizations
		if [ "$debug" = "on" ]; then
			optimize="-g"
		else
			optimize="-O2 -fomit-frame-pointer"
		fi
		extraflags=""
		case `uname -m` in
		  i?86)	if CheckNASM $subsystem; then
				extraflags="$extraflags -DUSE_ASMBLIT"
				nasmbuild="
%.o: %.asm
	\$(NASM) -o \$@ -f elf \$<
"
			fi
			;;
		  *)    ;;
		esac
		case $subsystem in
		  ggi)
			extrainc="-I/usr/local/include"
			extralibs="-L/usr/local/lib -lggi"
			;;
		  x11)
			extrainc=""
			extralibs="-L/usr/X11R6/lib -lX11 -lXext"
			if CheckXF86DGA; then
				echo \
"
* Detected XFree86 Direct Graphics Access extension (supported) "
#
				extraflags="$extraflags -DXFREE86_DGA"
				extralibs="$extralibs -lXxf86dga -lXxf86vm"
			fi
			;;
		  *)
			echo "Unknown subsystem '$subsystem' for $target"
			exit 3
			;;
		esac
		# Check for broken LinuxPPC pthreads
		if [ "`uname -m`" = "ppc" ]; then
			echo \
"
* Linux threads are currently broken under PPC (August '98) "
#
			extraflags="$extraflags -DFORK_HACK -D_REENTRANT"
			asmfiles=""
		elif CheckGLIBC; then
			echo \
"
* Compiling with pthreads -- you need compile your application with
   -D_REENTRANT and link with the pthread library (-lpthread)"
#
			extraflags="$extraflags -DSDL_USE_PTHREADS -D_REENTRANT"
			asmfiles=""
		else
			asmfiles="clone.S"
		fi
		sharedlib="libSDL${subsystem}.so.${version}"
		sharednam=`echo $sharedlib | sed 's/\.[0-9]$//'`
		shldopts="-shared -Wl,-soname,$sharednam"
		sharedrule="
$sharedlib: \$(SHOBJS)
	\$(CC) $shldopts -o \$@ \$(SHOBJS) $extralibs
"
		;;
	  sunos)
		# Allow GCC specific optimizations
		if [ "$debug" = "on" ]; then
			optimize="-g"
		else
			optimize="-O2 -fomit-frame-pointer"
		fi
		case $subsystem in
		  x11)
			extrainc="-I/usr/openwin/include"
			echo \
"
* Compiling with pthreads -- you need compile your application with
   -D_REENTRANT and link with the pthread library (-lpthread) "
#
			extraflags="-DSDL_USE_PTHREADS -D_REENTRANT"
			extralibs="-L/usr/openwin/lib -lX11 -lXext"
			;;
		  *)
			echo "Unknown subsystem '$subsystem' for $target"
			exit 3
			;;
		esac
		sharedlib="libSDL${subsystem}.so.${version}"
		shldopts="-G"
		sharedrule="
$sharedlib: \$(SHOBJS)
	ld $shldopts -o \$@ \$(SHOBJS) $extralibs
"
		;;
	  irix)
		if [ "$debug" = "on" ]; then
			optimize="-g"
		else
			optimize="-O2"
		fi
		case $subsystem in
		  x11)
			extrainc=""
			extraflags="-DSDL_USE_PTHREADS"
			extralibs="-lX11 -lXext -lpthread -laudio"
			;;
		  *)
			echo "Unknown subsystem '$subsystem' for $target"
			exit 3
			;;
		esac
		sharedlib="libSDL${subsystem}.so.${version}"
		shldopts="-shared"
		sharedrule="
$sharedlib: \$(SHOBJS)
	ld $shldopts -o \$@ \$(SHOBJS) $extralibs
"
		;;
	  osf1)
		if [ "$debug" = "on" ]; then
			optimize="-g"
		else
			optimize="-O2"
		fi
		case $subsystem in
		  x11)
			extrainc=""
			extraflags="-DSDL_USE_PTHREADS"
			extralibs="-lX11 -lXext -lpthreads -laudio"
			;;
		  *)
			echo "Unknown subsystem '$subsystem' for $target"
			exit 3
			;;
		esac
		sharedlib="libSDL${subsystem}.so.${version}"
		shldopts="-shared -expect_unresolved '*'"
		sharedrule="
$sharedlib: \$(SHOBJS)
	ld $shldopts -o \$@ \$(SHOBJS) $extralibs
"
		;;
	  beos)
		# Allow MWCC specific optimizations
		if [ "$debug" = "on" ]; then
			optimize="-g"
		else
			optimize="-O7"
		fi
		extraflags=""
		# Check for the limited linker
		prompt "Do you have the full Metrowerks linker? [y/N] "
		read full_linker
		if [ "$full_linker" != "y" -a "$full_linker" != "Y" ]; then
			echo \
"
The free Metrowerks linker has a limit of 64K code size -- some features off "
#
			extraflags="$extraflags -DLIMITED_LINKER"
		fi
		case $subsystem in
		  bwindow)
			;;
		  bscreen)
			;;
		  *)
			echo "Unknown subsystem '$subsystem' for $target"
			exit 3
			;;
		esac
		arch=`uname -m`
		case $arch in
		  BeMac)
			beobjs="/boot/develop/lib/ppc"
			;;
		  BePC)
			if CheckNASM $subsystem; then
				extraflags="$extraflags -DUSE_ASMBLIT"
				nasmbuild="
%.o: %.asm
	\$(NASM) -o \$@ -f win32 \$<
"
			fi
			beobjs="/boot/develop/lib/x86"
			;;
		  *)
			echo "Unknown Be architecture: $arch"
			exit 3
			;;
		esac
		extrainc=""
		extralibs="$beobjs/init_term_dyn.o $beobjs/glue-noinit.a"
		extralibs="$extralibs -lroot -lbe -lmedia"
		sharedlib="libSDL${subsystem}.so.${version}"
		shareddef="SDL.exp"
		if [ "$arch" = "BeMac" ]; then
			shldopts="-xms -f $shareddef"
		else
			shldopts="-xms"
		fi
		# How to build the shared library
		sharedrule="
$sharedlib: \$(SHOBJS)
	\$(CC) $shldopts -o \$@ \$(SHOBJS) $extralibs
"
		;;
	  win32)
		# Allow GCC specific optimizations
		if [ "$debug" = "on" ]; then
			optimize="-g"
		else
			optimize="-O2 -fomit-frame-pointer"
		fi
		extraflags=""
		if CheckNASM $subsystem; then
			extraflags="$extraflags -DUSE_ASMBLIT"
			nasmbuild="
%.o: %.asm
	\$(NASM) -o \$@ -f win32 \$<
"
		fi
		case $subsystem in
		  windib)
			extrainc="-I\$(TOP)/directx/include"
			extraflags=""
			extralibs="-luser32 -lgdi32 -lwinmm"
			sharedlib="SDL-dib.dll"
			;;
		  directx)
			extrainc="-I\$(TOP)/directx/include"
			extraflags=""
			extralibs="-luser32 -lgdi32 -lwinmm -L\$(TOP)/directx/lib -lddraw -ldsound -ldinput"
			sharedlib="SDL-dx5.dll"
			;;
		  *)
			echo "Unknown subsystem '$subsystem' for $target"
			exit 3
			;;
		esac
		shareddef="SDL.def"
		sharedrule="
$sharedlib: \$(SHOBJS)
	\$(CC) -mdll -o junk.tmp -Wl,--base-file,base.tmp \$(SHOBJS) $extralibs
	rm junk.tmp
	dlltool --dllname $sharedlib --base-file base.tmp --output-exp _temp.exp --def $shareddef
	rm base.tmp
	\$(CC) -mdll -o $sharedlib \$(SHOBJS) -Wl,_temp.exp $extralibs
	rm _temp.exp
"
		;;
	  *)	echo "$target is unsupported, sorry"
		exit 3
		;;
	esac
}

# Generate the library stub loader Makefile
#
# The following global variables must be set:
#
#	$target		The OS target (lowercase)
#	$stubinc	Header directories for building stub library
#	$stubflags	Compiler flags for building stub library
#	$*		The interface subsystems used
#
GenStubLoader()
{
	# Create the object directory
	subsystems="$*"
	if [ ! -d "./obj" ]; then
		mkdir "./obj" || exit 2
	fi
	if [ ! -d "./lib" ]; then
		mkdir "./lib" || exit 2
	fi

	# Link in all of the source files
	stobjects=""
	for file in src/stub/*.c src/stub/$target/*.c
	do  if [ -r $file ]; then
		rm -f ./obj/`stripdir $file`
		$makelink .. $file ./obj
		stobjects="$stobjects `basename $file .c`.o"
	    fi
	done
	for file in src/stub/*.h src/stub/*.decl src/stub/$target/*.decl
	do  if [ -r $file ]; then
		rm -f ./obj/`stripdir $file`
		$makelink .. $file ./obj
	    fi
	done

	# Generate the Makefile
	ranlib="`findexe ranlib`"
	if [ "$ranlib" = "" -o $target = "irix" ]; then
		ranlib=":"
	fi
	nasm="`findexe nasm`"
	if [ "$nasm" = "" ]; then
		nasm=":"
	fi
	cat >obj/Makefile <<__EOF__
#
# THIS MAKEFILE AUTOMATICALLY GENERATED BY `stripdir $0` -- DO NOT EDIT
#
DEBUG	= 
TOP	= ..
INCLUDES = -I\$(TOP)/include $stubinc
CFLAGS	= \$(DEBUG) \$(INCLUDES) $stubflags
STOBJS	= $stobjects
STLIB	= libSDL.a
RANLIB	= $ranlib

all: \$(STLIB) $sharedlibs

\$(STLIB): \$(STOBJS)
	for dir in $subsystems; do \\
		\$(MAKE) -C \$\$dir CC=\$(CC); \\
	done
	ar crv \$@ \$(STOBJS)
	\$(RANLIB) \$@

clean:
	rm -f \$(STOBJS) \$(STLIB)
__EOF__
}

# Generate an interface subsystem Makefile
#
# The following global variables must be set:
#
#	$version	The SDL library version
#	$target		The OS target (lowercase)
#	$1		The interface subsystem directory
#	$optimize	The compiler optimization flags
#	$extrainc	Header directories for building shared library
#	$extraflags	Compiler flags for building shared library
#	$sharedlib	The name of the shared library object file
#	$sharedrule	Makefile rule used to build shared library
#
GenSubsystem()
{
	# Create the subsystem object directory
	subsystem=$1
	if [ ! -d "./obj/$subsystem" ]; then
		mkdir "./obj/$subsystem"
	fi

	# Link in all of the source files
	shobjects=""
	for file in src/*.c src/$target/*.c src/$target/$subsystem/*.c
	do  if [ -r $file ]; then
		rm -f ./obj/$subsystem/`stripdir $file`
		$makelink ../.. $file ./obj/$subsystem
		shobjects="$shobjects `basename $file .c`.o"
	    fi
	done
	for file in src/$target/*.cc src/$target/$subsystem/*.cc
	do  if [ -r $file ]; then
		rm -f ./obj/$subsystem/`stripdir $file`
		$makelink ../.. $file ./obj/$subsystem
		shobjects="$shobjects `basename $file .cc`.o"
	    fi
	done
	if [ "$nasmbuild" != "" ]; then
		for file in src/hermes/*.h
		do  if [ -r "$file" ]; then
			rm -f ./obj/$subsystem/`stripdir $file`
			$makelink ../.. $file ./obj/$subsystem
		    fi
		done
		for file in src/hermes/*.asm
		do if [ -r $file ]; then
			rm -f ./obj/$subsystem/`stripdir $file`
			$makelink ../.. $file ./obj/$subsystem
			shobjects="$shobjects `basename $file .asm`.o"
 		   fi
		done
	fi
	for file in $asmfiles
	do file="src/$target/$file"
	   if [ -r $file ]; then
		rm -f ./obj/$subsystem/`stripdir $file`
		$makelink ../.. $file ./obj/$subsystem
		shobjects="$shobjects `basename $file .S`.o"
 	   fi
	done
	for file in src/*.h src/$target/*.h src/$target/$subsystem/*.h
	do  if [ -r "$file" ]; then
		rm -f ./obj/$subsystem/`stripdir $file`
		$makelink ../.. $file ./obj/$subsystem
	    fi
	done
	for file in src/$target/exports/SDL.*
	do if [ -r $file ]; then
		rm -f ./obj/$subsystem/`stripdir $file`
		$makelink ../.. $file ./obj/$subsystem
   	fi
	done

	# Generate the Makefile
	cat >obj/$subsystem/Makefile <<__EOF__
#
# THIS MAKEFILE AUTOMATICALLY GENERATED BY `stripdir $0` -- DO NOT EDIT
#
CXX = \$(CC)
DEBUG	= $optimize
TOP	= ../..
INCLUDES = -I\$(TOP)/include $extrainc
CFLAGS	= \$(DEBUG) \$(INCLUDES) \$(STATIC) -D_BUILDING_SDL $extraflags
CXXFLAGS = \$(CFLAGS)
SHOBJS	= $shobjects
RANLIB	= $ranlib
NASM	= $nasm

all: $sharedlib

$nasmbuild

static: clean
	\$(MAKE) libSDL.a STATIC=-D_SDL_STATIC_LIB

libSDL.a: \$(SHOBJS)
	ar crv \$@ \$(SHOBJS)
	\$(RANLIB) \$@

$sharedrule

clean:
	rm -f \$(SHOBJS)
__EOF__
}

# Figure out target compilation
if [ "$1" = "" ]; then
	target=`uname | tr '[A-Z]' '[a-z]'`
else
	target=$1
fi
copy_link()
{
	cp $2 $3
}
link_link()
{
	ln -s $1/$2 $3
}
if [ $target = "cygwin32_95" -o $target = "cygwin32_nt" ]; then
	target="win32"
	makelink=copy_link
else
	makelink=link_link
fi

# Make sure we're in the right directory
if [ ! -d src ]; then
	echo "Please run this script from the toplevel directory"
	exit 1
fi

# Generate the main Makefile config
sharedlibs=""
case $target in
  linux)
	stubinc=""
	stubflags=""
	subsystems=`SelectSubsystems X11`
	#subsystems=`SelectSubsystems X11 GGI`
	;;
  sunos)
	# Create links to Linux source code
	for link in SDL_mutex.c SDL_thread.c SDL_timer.c x11
	do if [ ! -h src/sunos/$link ]; then
		ln -s ../linux/$link src/sunos/$link
	   fi
	done
	stubinc="-I/usr/openwin/include"
	stubflags=""
	subsystems=`SelectSubsystems X11`
	;;
  irix)
	# Create links to Linux source code
	for link in SDL_mutex.c SDL_thread.c SDL_timer.c x11
	do if [ ! -h src/irix/$link ]; then
		ln -s ../linux/$link src/irix/$link
	   fi
	done
	stubinc=""
	stubflags=""
	subsystems=`SelectSubsystems X11`
	;;
  osf1)
	# Create links to Linux source code
	for link in SDL_timer.c x11
	do if [ ! -h src/osf1/$link ]; then
		ln -s ../linux/$link src/osf1/$link
	   fi
	done
	# Create links to SunOS source code
	for link in SDL_sysaudio.c
	do if [ ! -h src/osf1/$link ]; then
		ln -s ../sunos/$link src/osf1/$link
	   fi
	done
	stubinc=""
	stubflags=""
	subsystems=`SelectSubsystems X11`
	;;
  win32)
	stubinc="-I../directx/include"
	stubflags=""
	if CheckWinGCC; then
		:
	else
		echo "
Couldn't compile a simple windows program.
 -- is your Win32 C compiler first in your execution path?
		"
		exit 1
	fi
	if CheckDirectX; then
		subsystems=`SelectSubsystems WinDIB DirectX`
	else
		prompt \
"Couldn't find DirectX headers -- will you install them?  [y/N] "
		read have_headers
		if [ "$have_headers" = "y" -o "$have_headers" = "Y" ]; then
			echo \
"
* Read the directx/README file for header installation instructions
"
			exit 1
		else
			stubflags="$stubflags -DNO_DIRECTX_HEADERS"
		fi
		subsystems=`SelectSubsystems WinDIB`
	fi
	;;
  beos)
	stubinc=""
	stubflags=""
	subsystems=`SelectSubsystems BWindow`
	;;
  *)
	echo "$target is unsupported, sorry"
	exit 3
	;;
esac
echo "Linking sources for $target..."
GenStubLoader $subsystems
for subsystem in $subsystems
do  GenMakeflags $subsystem
    GenSubsystem $subsystem
    sharedlibs="$sharedlibs obj/$subsystem/$sharedlib"
done
cat >make.gen <<__EOF__
#
# THIS MAKEFILE AUTOMATICALLY GENERATED BY `stripdir $0` -- DO NOT EDIT
#

all: lib/libSDL.a

lib/libSDL.a: obj/libSDL.a $sharedlibs
	@for lib in $sharedlibs; do \\
	  echo "\$\$lib -> lib/\`basename \$\$lib\`"; \\
	  cp \$\$lib lib ; \\
	  lib=\`basename \$\$lib\`; \\
	  link=\`echo \$\$lib | sed 's/\\(libSDL.*\\.so\\..\\..\\).*/\\1/'\`; \\
	  if [ ! -r lib/\$\$link ]; then \\
	    echo "lib/\$\$lib <- lib/\$\$link"; \\
	    ln -s \$\$lib lib/\$\$link; \\
	  fi; \\
	done
	@echo "obj/libSDL.a -> lib/libSDL.a"
	@cp obj/libSDL.a lib

obj/libSDL.a:
	\$(MAKE) -C obj

__EOF__
mv make.gen make.lib

# Set up links for the system dependent Makefile fragments
echo $target >.target

# Done.
echo ""
echo "Please type 'make' to build SDL $version for $target"
echo ""
