/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_bmp.c,v 1.6 1998/10/07 16:56:37 slouken Exp $";
#endif
/* 
   Code to load and save surfaces in Windows BMP format.

   Why support BMP format?  Well, it's a native format for Windows, and
   most image processing programs can read and write it.  It would be nice
   to be able to have at least one image format that we can natively load
   and save, and since PNG is so complex that it would bloat the library,
   BMP is a good alternative. 

   This code currently supports Win32 DIBs in uncompressed 8 and 24 bpp.
*/

#include <string.h>

#include "SDL_error.h"
#include "SDL_video.h"
#include "SDL_endian.h"

/* Compression encodings for BMP files */
#ifndef BI_RGB
#define BI_RGB		0
#define BI_RLE8		1
#define BI_RLE4		2
#define BI_BITFIELD	3
#endif

SDL_Surface * SDL_LoadBMP (const char *file)
{
	FILE  *bmp;
	Uint32 i, pad;
	int    c;
	SDL_Surface *surface;
	Uint32 Rmask;
	Uint32 Gmask;
	Uint32 Bmask;
	SDL_Palette *palette;
	Uint8 *bits;
	int    ExpandBMP;

	/* The Win32 BMP file header (14 bytes) */
	char   magic[2];
	Uint32 bfSize;
	Uint16 bfReserved1;
	Uint16 bfReserved2;
	Sint32 bfOffBits;

	/* The Win32 BITMAPINFOHEADER struct (40 bytes) */
	Uint32 biSize;
	Sint32 biWidth;
	Sint32 biHeight;
	Uint16 biPlanes;
	Uint16 biBitCount;
	Uint32 biCompression;
	Uint32 biSizeImage;
	Sint32 biXPelsPerMeter;
	Sint32 biYPelsPerMeter;
	Uint32 biClrUsed;
	Uint32 biClrImportant;

	bmp = fopen(file, "rb");
	if ( bmp == NULL ) {
		SDL_SetError("Couldn't open %s for reading", file);
		return(NULL);
	}

	/* Read in the BMP file header */
	SDL_ClearError();
	if ( fread(magic, 1, 2, bmp) != 2 ) {
		SDL_SetError("Error reading from %s", file);
		fclose(bmp);
		return(NULL);
	}
	if ( strncmp(magic, "BM", 2) != 0 ) {
		SDL_SetError("%s is not a Windows BMP file", file);
		fclose(bmp);
		return(NULL);
	}
	SDL_CalculateEndian();
	bfSize		= SDL_ReadLE32(bmp);
	bfReserved1	= SDL_ReadLE16(bmp);
	bfReserved2	= SDL_ReadLE16(bmp);
	bfOffBits	= SDL_ReadLE32(bmp);

	/* Read the Win32 BITMAPINFOHEADER */
	biSize		= SDL_ReadLE32(bmp);
	if ( biSize == 12 ) {
		biWidth		= (Uint32)SDL_ReadLE16(bmp);
		biHeight	= (Uint32)SDL_ReadLE16(bmp);
		biPlanes	= SDL_ReadLE16(bmp);
		biBitCount	= SDL_ReadLE16(bmp);
		biCompression	= BI_RGB;
		biSizeImage	= 0;
		biXPelsPerMeter	= 0;
		biYPelsPerMeter	= 0;
		biClrUsed	= 0;
		biClrImportant	= 0;
	} else {
		biWidth		= SDL_ReadLE32(bmp);
		biHeight	= SDL_ReadLE32(bmp);
		biPlanes	= SDL_ReadLE16(bmp);
		biBitCount	= SDL_ReadLE16(bmp);
		biCompression	= SDL_ReadLE32(bmp);
		biSizeImage	= SDL_ReadLE32(bmp);
		biXPelsPerMeter	= SDL_ReadLE32(bmp);
		biYPelsPerMeter	= SDL_ReadLE32(bmp);
		biClrUsed	= SDL_ReadLE32(bmp);
		biClrImportant	= SDL_ReadLE32(bmp);
	}

	/* Check for read error */
	if ( strcmp(SDL_GetError(), "") != 0 ) {
		fclose(bmp);
		return(NULL);
	}

	/* Expand 1 and 4 bit bitmaps to 8 bits per pixel */
	switch (biBitCount) {
		case 1:
		case 4:
			ExpandBMP = biBitCount;
			biBitCount = 8;
			break;
		default:
			ExpandBMP = 0;
			break;
	}

	/* We don't support any BMP compression right now */
	Rmask = Gmask = Bmask = 0;
	switch (biCompression) {
		case BI_RGB:
			switch (biBitCount) {
				case 15:
				case 16:
					Rmask = 0x7C00;
					Gmask = 0x03E0;
					Bmask = 0x001F;
					break;
				case 24:
				case 32:
					Bmask = 0x00FF0000;
					Gmask = 0x0000FF00;
					Rmask = 0x000000FF;
					break;
				default:
					break;
			}
			break;
		case BI_BITFIELD:
			switch (biBitCount) {
				case 15:
				case 16:
				case 32:
					Rmask = SDL_ReadLE32(bmp);
					Gmask = SDL_ReadLE32(bmp);
					Bmask = SDL_ReadLE32(bmp);
					break;
				default:
					break;
			}
			break;
		default:
			SDL_SetError("Compressed BMP files not supported");
			fclose(bmp);
			return(NULL);
	}
  
  	/* Create a compatible surface, note that the colors are RGB ordered */
  	surface = SDL_AllocSurface(SDL_SWSURFACE,
			biWidth, biHeight, biBitCount, Rmask, Gmask, Bmask, 0);
	if ( surface == NULL ) {
		fclose(bmp);
		return(NULL);
	}

	/* Load the palette, if any */
	palette = (surface->format)->palette;
	if ( palette ) {
		if ( biClrUsed == 0 ) {
			biClrUsed = 1 << biBitCount;
		}
		if ( biSize == 12 ) {
			for ( i = 0; i < biClrUsed; ++i ) {
				palette->colors[i].b = (Uint8)fgetc(bmp);
				palette->colors[i].g = (Uint8)fgetc(bmp);
				palette->colors[i].r = (Uint8)fgetc(bmp);
				palette->colors[i].unused = 0;
			}	
		} else {
			for ( i = 0; i < biClrUsed; ++i ) {
				palette->colors[i].b = (Uint8)fgetc(bmp);
				palette->colors[i].g = (Uint8)fgetc(bmp);
				palette->colors[i].r = (Uint8)fgetc(bmp);
				palette->colors[i].unused = (Uint8)fgetc(bmp);
			}	
		}
		palette->ncolors = biClrUsed;
	}

	/* Read the surface pixels.  Note that the bmp image is upside down */
	fseek(bmp, (long)bfOffBits, SEEK_SET);
	bits = (Uint8 *)surface->pixels+(surface->h*surface->pitch);
	switch (ExpandBMP) {
		case 1:
		case 4:
			pad  = (((surface->pitch/(8/ExpandBMP))%4) ?
				(4-((surface->pitch/(8/ExpandBMP))%4)) : 0);
			break;
		default:
			pad  = ((surface->pitch%4) ?
					(4-(surface->pitch%4)) : 0);
			break;
	}
	while ( bits > (Uint8 *)surface->pixels ) {
		bits -= surface->pitch;
		switch (ExpandBMP) {
			case 1:
			case 4: {
			Uint8 pixel = 0;
			int   shift = (8-ExpandBMP);
			for ( i=0; i<surface->w; ++i ) {
				if ( i%(8/ExpandBMP) == 0 ) {
					c = fgetc(bmp);
					if ( c == EOF ) {
						SDL_SetError(
					"Error reading from BMP file");
						fclose(bmp);
						SDL_FreeSurface(surface);
						return(NULL);
					}
					pixel = (Uint8)c;
				}
				*(bits+i) = (pixel>>shift);
				pixel <<= ExpandBMP;
			} }
			break;

			default:
			if (fread(bits,1,surface->pitch,bmp) != surface->pitch){
				SDL_SetError("Error reading from BMP file");
				fclose(bmp);
				SDL_FreeSurface(surface);
				return(NULL);
			}
			break;
		}
		for ( i=0; i<pad; ++i )
			fgetc(bmp);
	}
	fclose(bmp);
	return(surface);
}

int SDL_SaveBMP (SDL_Surface *saveme, const char *file)
{
	FILE  *bmp;
	int i, pad;
	SDL_Surface *surface;
	Uint8 *bits;

	/* The Win32 BMP file header (14 bytes) */
	char   magic[2] = { 'B', 'M' };
	Uint32 bfSize;
	Uint16 bfReserved1;
	Uint16 bfReserved2;
	Sint32 bfOffBits;

	/* The Win32 BITMAPINFOHEADER struct (40 bytes) */
	Uint32 biSize;
	Sint32 biWidth;
	Sint32 biHeight;
	Uint16 biPlanes;
	Uint16 biBitCount;
	Uint32 biCompression;
	Uint32 biSizeImage;
	Sint32 biXPelsPerMeter;
	Sint32 biYPelsPerMeter;
	Uint32 biClrUsed;
	Uint32 biClrImportant;

	SDL_ClearError();
	if ( saveme->format->palette ) {
		if ( saveme->format->BitsPerPixel != 8 ) {
			SDL_SetError("%d bpp BMP files not supported",
						saveme->format->BitsPerPixel);
			return(-1);
		}
		surface = saveme;
	} else
	if ( saveme->format->BitsPerPixel != 24 ) {
		SDL_PixelFormat *format;

		/* Convert to 24 bits per pixel */
		surface = SDL_AllocSurface(SDL_SWSURFACE,
					saveme->w, saveme->h, 24,
					0x00FF0000, 0x0000FF00, 0x000000FF, 0);
		format = saveme->dstfmt;
		if ( SDL_MapSurface(saveme, surface->format) < 0 )
			return(-1);
		SDL_BlitSurface(saveme, NULL, surface, NULL);
		SDL_MapSurface(saveme, format);
	} else {
		surface = saveme;
	}

	bmp = fopen(file, "wb");
	if ( bmp == NULL ) {
		SDL_SetError("Couldn't open %s for writing", file);
		return(-1);
	}

	/* Set the BMP file header values */
	bfSize = 0;		 /* We'll write this when we're done */
	bfReserved1 = 0;
	bfReserved2 = 0;
	bfOffBits = 0;		/* We'll write this when we're done */

	/* Write the BMP file header values */
	SDL_CalculateEndian();
	fwrite(magic, 2, 1, bmp);
	SDL_WriteLE32(bmp, bfSize);
	SDL_WriteLE16(bmp, bfReserved1);
	SDL_WriteLE16(bmp, bfReserved2);
	SDL_WriteLE32(bmp, bfOffBits);

	/* Set the BMP info values */
	biSize = 40;
	biWidth = surface->w;
	biHeight = surface->h;
	biPlanes = 1;
	biBitCount = surface->format->BitsPerPixel;
	biCompression = BI_RGB;
	biSizeImage = surface->h*surface->pitch;
	biXPelsPerMeter = 0;
	biYPelsPerMeter = 0;
	if ( surface->format->palette ) {
		biClrUsed = surface->format->palette->ncolors;
	} else {
		biClrUsed = 0;
	}
	biClrImportant = 0;

	/* Write the BMP info values */
	SDL_WriteLE32(bmp, biSize);
	SDL_WriteLE32(bmp, biWidth);
	SDL_WriteLE32(bmp, biHeight);
	SDL_WriteLE16(bmp, biPlanes);
	SDL_WriteLE16(bmp, biBitCount);
	SDL_WriteLE32(bmp, biCompression);
	SDL_WriteLE32(bmp, biSizeImage);
	SDL_WriteLE32(bmp, biXPelsPerMeter);
	SDL_WriteLE32(bmp, biYPelsPerMeter);
	SDL_WriteLE32(bmp, biClrUsed);
	SDL_WriteLE32(bmp, biClrImportant);

	/* Write the palette (in BGR color order) */
	if ( surface->format->palette ) {
		SDL_Color *colors;
		int       ncolors;

		colors = surface->format->palette->colors;
		ncolors = surface->format->palette->ncolors;
		for ( i=0; i<ncolors; ++i ) {
			fputc(colors[i].b, bmp);
			fputc(colors[i].g, bmp);
			fputc(colors[i].r, bmp);
			fputc(colors[i].unused, bmp);
		}
	}

	/* Write the bitmap offset */
	bfOffBits = ftell(bmp);
	fseek(bmp, 10, SEEK_SET);
	SDL_WriteLE32(bmp, bfOffBits);
	fseek(bmp, bfOffBits, SEEK_SET);

	/* Write the bitmap image upside down */
	bits = (Uint8 *)surface->pixels+(surface->h*surface->pitch);
	pad  = ((surface->pitch%4) ? (4-(surface->pitch%4)) : 0);
	while ( bits > (Uint8 *)surface->pixels ) {
		bits -= surface->pitch;
		if ( fwrite(bits, 1, surface->pitch, bmp) != surface->pitch ) {
			SDL_SetError("Error writing to datastream");
			break;
		}
		for ( i=0; i<pad; ++i )
			fputc(0, bmp);
	}

	/* Write the BMP file size */
	bfSize = ftell(bmp);
	fseek(bmp, 2, SEEK_SET);
	SDL_WriteLE32(bmp, bfSize);
	fseek(bmp, bfSize, SEEK_SET);

	/* Close it up.. */
	if ( fclose(bmp) == EOF ) {
		SDL_SetError("Error writing to datastream");
	}
	if ( surface != saveme ) {
		SDL_FreeSurface(surface);
	}
	return((strcmp(SDL_GetError(), "") == 0) ? 0 : -1);
}
