/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_video.c,v 1.10 1998/10/08 08:02:40 slouken Exp $";
#endif

/* The high-level video driver subsystem */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SDL_error.h"
#include "SDL_video.h"
#include "SDL_events.h"
#include "SDL_mutex.h"
#include "SDL_sysvideo.h"
#include "SDL_sysevents.h"
#include "SDL_video_c.h"
#include "SDL_pixels_c.h"
#include "SDL_events_c.h"
#include "SDL_cursor_c.h"

/* Stub functions to stick into our hardware caps */
static int SDL_NoHWAlloc(SDL_Surface *surface);
static int SDL_NoHWAccel(SDL_Surface *surface);

/* Global state flags */
struct WMcaps SDL_WMCaps = {
        NULL,				/* FullScreen */
        NULL,				/* SetIcon */
        NULL,				/* SetCaption */
	NULL				/* GetWMInfo */
};
struct HWcaps SDL_HWCaps = {
	SDL_NoHWAlloc,			/* AllocHWSurface */
	SDL_NoHWAccel,			/* CheckHWAccel */
	NULL,				/* SetHWColorKey */
	NULL,				/* SetHWAlpha */
	NULL,				/* LockHWSurface */
	NULL,				/* UlockHWSurface */
	NULL				/* FreeHWSurface */
};
SDL_Surface *SDL_PublicSurface = NULL;
SDL_Surface *SDL_VideoSurface  = NULL;
SDL_Surface *SDL_ShadowSurface = NULL;
SDL_Surface *SDL_FadedSurface  = NULL;
void (*SDL_SYS_UpdateRects)
		(SDL_Surface *screen, int numrects, SDL_Rect *rects) = NULL;

/* Places to store title and icon text for the app */
static char *wm_title = NULL;
static char *wm_icon  = NULL;
static int offset_x, offset_y;

/* Various local functions */
int SDL_VideoInit(void);
static void SDL_AllocShadowSurface(Uint8 depth);
extern int  SDL_MapShadowSurface(void);
static void SDL_FreeShadowSurface(void);
void SDL_VideoQuit(void);

/*
 * Initialize the video and event subsystems -- determine native pixel format
 */
int SDL_VideoInit (void)
{
	SDL_PixelFormat vformat;
	Uint32 video_flags;

	/* Initialize the video subsystem, get the "best" pixel format */
	memset(&vformat, 0, sizeof(vformat));
	if ( SDL_SYS_VideoInit(&vformat) < 0 ) {
		SDL_VideoQuit();
		return(-1);
	}

	/* Create a zero sized video surface of the appropriate format */
	video_flags = SDL_SWSURFACE;
#ifdef THREADED_EVENTS
	video_flags |= SDL_THREADSAFE;
#endif
	SDL_VideoSurface = SDL_AllocSurface(video_flags, 0, 0,
				vformat.BitsPerPixel,
				vformat.Rmask, vformat.Gmask, vformat.Bmask, 0);
	if ( SDL_VideoSurface == NULL ) {
		SDL_VideoQuit();
		return(-1);
	}
	SDL_PublicSurface = NULL;	/* Until SDL_SetVideoMode() */

	/* If we have a palettized surface, create a default palette */
	if ( SDL_VideoSurface->format->palette ) {
		SDL_DitherColors(SDL_VideoSurface->format->palette->colors,
					SDL_VideoSurface->format->BitsPerPixel);
		SDL_SYS_SetColors(SDL_VideoSurface,
				0, SDL_VideoSurface->format->palette->ncolors);
	}

	/* Start the event loop */
	if ( SDL_StartEventLoop() < 0 ) {
		SDL_VideoQuit();
		return(-1);
	}

	/* We're ready to go! */
	return(0);
}

/*
 * Get the current pixel format of the display surface.
 */
SDL_PixelFormat * SDL_GetDisplayFormat (SDL_PixelFormat *format)
{
	memcpy(format, SDL_VideoSurface->format, sizeof(*format));
	return(format);
}

/*
 * Return a pointer to an array of available screen dimensions for the
 * given format, sorted largest to smallest.  Returns NULL if there are
 * no dimensions available for a particular format, or (SDL_Rect **)-1
 * if any dimension is okay for the given format.  If 'format' is NULL,
 * the mode list will be for the format returned by SDL_GetPixelFormat()
 */
SDL_Rect ** SDL_ListModes (SDL_PixelFormat *format, Uint32 flags)
{
	if ( format == NULL ) {
		format = SDL_VideoSurface->format;
	}
	return(SDL_SYS_ListModes(SDL_VideoSurface, format, flags));
}

/*
 * Check to see if a particular video mode is supported.
 * It returns 0 if the requested mode is not supported under any bit depth,
 * or returns the bits-per-pixel of the closest available mode with the
 * given width and height.  If this bits-per-pixel is different from the
 * one used when setting the video mode, SDL_SetVideoMode() will succeed,
 * but will emulate the requested bits-per-pixel with a shadow surface.
 */
int SDL_VideoModeOK (int width, int height, int bpp, Uint32 flags) 
{
	Uint8 bit_depths[4][7] = {
		/* 8 bit closest depth ordering */
		{ 0, 8, 16, 15, 32, 24, 0 },
		/* 15,16 bit closest depth ordering */
		{ 0, 16, 15, 32, 24, 8, 0 },
		/* 24 bit closest depth ordering */
		{ 0, 24, 32, 16, 15, 8, 0 },
		/* 32 bit closest depth ordering */
		{ 0, 32, 16, 15, 24, 8, 0 }
	};
	int table, b, i;
	int supported;
	SDL_PixelFormat format;
	SDL_Rect **sizes;

	/* Currently 1 and 4 bpp are not supported */
	if ( bpp < 8 ) {
		return(0);
	}

	/* Search through the list valid of modes */
	memset(&format, 0, sizeof(format));
	supported = 0;
	table = ((bpp+7)/8)-1;
	bit_depths[table][0] = bpp;
	for ( b = 0; !supported && bit_depths[table][b]; ++b ) {
		format.BitsPerPixel = bit_depths[table][b];
		sizes = SDL_ListModes(&format, flags);
		if ( sizes == (SDL_Rect **)0 ) {
			/* No sizes supported at this bit-depth */
			continue;
		} else 
		if ( sizes == (SDL_Rect **)-1 ) {
			/* Any size supported at this bit-depth */
			supported = 1;
			continue;
		} else
		for ( i=0; sizes[i]; ++i ) {
			if ((sizes[i]->w == width) && (sizes[i]->h == height)) {
				supported = 1;
				break;
			}
		}
	}
	if ( supported ) {
		--b;
		return(bit_depths[table][b]);
	} else {
		return(0);
	}
}

/*
 * Get the closest non-emulated video mode to the one requested
 */
static int SDL_GetVideoMode (int *w, int *h, int *BitsPerPixel, Uint32 flags)
{
	SDL_Rect **modes;
	int i, okay;
	Uint8 native_bpp;

	/* Try the original video mode */
	native_bpp = SDL_VideoModeOK(*w, *h, *BitsPerPixel, flags);
	if ( native_bpp == *BitsPerPixel ) {;
		return(1);
	}
	if ( native_bpp > 0 ) {
		*BitsPerPixel = native_bpp;
		return(1);
	}

	/* Enumerate the native video modes, looking for a good size */
	native_bpp = SDL_VideoSurface->format->BitsPerPixel;
	okay = 1;
	modes = SDL_ListModes(SDL_VideoSurface->format, flags);

	/* Check to see if the requested mode is available */
	if ( okay && (modes == (SDL_Rect **)-1) ) {
		/* Uh oh, supports any size, but bit depth failed above (??) */
		SDL_SetError("No video modes supported. (??)");
		okay = 0;
	}
	if ( okay && (modes == (SDL_Rect **)0) ) {
		/* Uh oh, no modes supported at native format (??) */
		SDL_SetError("No video modes supported. (??)");
		okay = 0;
	}
	if ( okay && ((modes[0]->w < *w) || (modes[0]->h < *h)) ) {
		SDL_SetError(
		"Biggest mode (%dx%d) is smaller than desired (%dx%d)",
					modes[0]->w, modes[0]->h, *w, *h);
		okay = 0;
	}
	if ( ! okay ) {
		return(0);
	}

	/* Look for the closest sized mode at native bitdepth */
	for ( i=0; modes[i] && (modes[i]->w > *w) && (modes[i]->h > *h); ++i ) {
		/* keep looking */;
	}
	if ( ! modes[i] || (modes[i]->w < *w) || (modes[i]->h < *h) ) {
		--i;	/* We went too far */
	}

	/* We found the smallest mode >= width x height */
	*w = modes[i]->w;
	*h = modes[i]->h;
	*BitsPerPixel = native_bpp;
	return(1);
}

/* This should probably go somewhere else -- like SDL_surface.c */
static int SDL_ClearSurface(SDL_Surface *surface)
{
	Uint32 black;
	int y;
	Uint8 *row;
	int bytewidth;

	/* Note that memset() only works with black */
	black = SDL_MapRGB(surface->format, 0, 0, 0);
	if ( SDL_LockSurface(surface) == 0 ) {
		bytewidth = surface->w*surface->format->BytesPerPixel;
		row = (Uint8 *)surface->pixels;
		for ( y=surface->h; y; --y ) {
			memset(row, black, bytewidth);
			row += surface->pitch;
		}
		SDL_UnlockSurface(surface);
	}
	SDL_UpdateRect(surface, 0, 0, 0, 0);
}

/*
 * Set the requested video mode, allocating a shadow buffer if necessary.
 */
SDL_Surface * SDL_SetVideoMode (int width, int height, int bpp, Uint32 flags)
{
	SDL_Surface *mode;
	int video_w;
	int video_h;
	int video_bpp;

	/* Default to the current video bpp */
	if ( bpp == 0 ) {
		flags |= SDL_ANYFORMAT;
		bpp = SDL_VideoSurface->format->BitsPerPixel;
	}

	/* Get a good video mode, the closest one possible */
	video_w = width;
	video_h = height;
	video_bpp = bpp;
	if ( ! SDL_GetVideoMode(&video_w, &video_h, &video_bpp, flags) ) {
		return(NULL);
	}

	/* Check the requested flags */
	if ( video_bpp > 8 ) {
		flags &= ~SDL_HWPALETTE;
	}

	/* Clean up any previous video mode */
	if ( SDL_PublicSurface != NULL ) {
		SDL_PublicSurface = NULL;
	}
	SDL_FreeShadowSurface();

	/* Try to set the video mode, along with offset and clipping */
	mode = SDL_SYS_SetVideoMode(SDL_VideoSurface, video_w, video_h,
							video_bpp, flags);
	if ( mode != NULL ) {
		if ( (mode->w < width) || (mode->h < height) ) {
			SDL_SetError("Video mode smaller than requested");
			return(NULL);
		}
		mode->offset = 0;
		SDL_ClearSurface(mode);
		offset_x = (mode->w-width)/2;
		offset_y = (mode->h-height)/2;
		mode->offset = offset_y*mode->pitch +
				offset_x*mode->format->BytesPerPixel;
#ifdef VIDEO_DEBUG
  fprintf(stderr,
	"Requested mode: %dx%dx%d, obtained mode %dx%dx%d (offset %d)\n",
		width, height, bpp,
		mode->w, mode->h, mode->format->BitsPerPixel, mode->offset);
#endif
		mode->w = width;
		mode->h = height;
		SDL_SetClipping(mode, 0, 0, 0, 0);
	} else {
		return(NULL);
	}
	if ( mode != SDL_VideoSurface ) {
		SDL_FreeSurface(SDL_VideoSurface);
		SDL_VideoSurface = mode;
	}

	/* If we have a palettized surface, create a default palette */
	if ( SDL_VideoSurface->format->palette ) {
		SDL_DitherColors(SDL_VideoSurface->format->palette->colors,
					SDL_VideoSurface->format->BitsPerPixel);
		SDL_SYS_SetColors(SDL_VideoSurface,
				0, SDL_VideoSurface->format->palette->ncolors);
	}

	/* Create a shadow surface if necessary */
	/* There are two conditions under which we create a shadow surface:
		1.  We need a particular bits-per-pixel that we didn't get.
		2.  We need a hardware palette and didn't get one.
	*/
	if ( ( !(flags&SDL_ANYFORMAT) &&
			(SDL_VideoSurface->format->BitsPerPixel != bpp)) ||
	     (  (flags&SDL_HWPALETTE) && 
				!(SDL_VideoSurface->flags&SDL_HWPALETTE)) ) {
		SDL_AllocShadowSurface(bpp);
		if ( SDL_ShadowSurface == NULL ) {
			SDL_SetError("Couldn't create shadow surface");
			return(NULL);
		}
		SDL_PublicSurface = SDL_ShadowSurface;
	} else {
		SDL_PublicSurface = SDL_VideoSurface;
	}

	/* Make the surface thread-safe if so desired */
	if ( (flags & SDL_THREADSAFE) && 
			! (SDL_PublicSurface->flags & SDL_THREADSAFE) ) {
		SDL_PublicSurface->mutex = SDL_CreateMutex();
		if ( SDL_PublicSurface->mutex != NULL ) {
			SDL_PublicSurface->flags |= SDL_THREADSAFE;
		}
	}

	/* The video mode has been set -- create our cursor */
	SDL_CursorInit();

	/* We're done! */
	return(SDL_PublicSurface);
}

/* 
 * Convert a surface into the video pixel format.
 */
SDL_Surface * SDL_DisplayFormat (SDL_Surface *surface)
{
	SDL_Surface *vidsurface;

	/* Don't convert to the format of a faded surface */
	if ( SDL_ShadowSurface != NULL ) {
		vidsurface = SDL_ShadowSurface;
	} else {
		vidsurface = SDL_VideoSurface;
	}
	/* Only make surface thread-safe if original is thread-safe */
	if ( surface->flags & SDL_THREADSAFE ) {
		return(SDL_ConvertSurface(surface, vidsurface->format,
					vidsurface->flags | SDL_THREADSAFE));
	} else {
		return(SDL_ConvertSurface(surface, vidsurface->format,
					vidsurface->flags & ~SDL_THREADSAFE));
	}
}

/* Variable to save the shadow mapping during blits */
struct surface_map {
	int identity;
	SDL_PixelFormat *dstfmt;
	SDL_blit hw_blit;
	SDL_blit sw_blit;
	Uint8    *palmap;
	struct private_hwdata *hwaccel_data;
	struct private_swdata *swaccel_data;
} shadow_map, user_map;

static void SDL_EmptySurfaceMap(struct surface_map *map)
{
	memset(map, 0, sizeof(*map));
}
static void SDL_FreeSurfaceMap(struct surface_map *map)
{
	if ( map->palmap != NULL ) {
		free(map->palmap);
		map->palmap = NULL;
	}
}
static void SDL_SaveSurfaceMap(struct surface_map *map, SDL_Surface *surface)
{
	map->identity = surface->identity;
	map->dstfmt = surface->dstfmt;
	map->hw_blit = surface->hw_blit;
	map->sw_blit = surface->sw_blit;
	map->palmap = surface->palmap;
	map->hwaccel_data = surface->hwaccel_data;
	map->swaccel_data = surface->swaccel_data;
}
static void SDL_RestoreSurfaceMap(struct surface_map *map, SDL_Surface *surface)
{
	surface->identity = map->identity;
	surface->dstfmt = map->dstfmt;
	surface->hw_blit = map->hw_blit;
	surface->sw_blit = map->sw_blit;
	surface->palmap = map->palmap;
	surface->hwaccel_data = map->hwaccel_data;
	surface->swaccel_data = map->swaccel_data;
}

int SDL_MapShadowSurface(void)
{
	int retval;

	/* Save current mapping and clear it for the new map */
	SDL_SaveSurfaceMap(&user_map, SDL_ShadowSurface);
	SDL_ShadowSurface->palmap = NULL;

	/* Map to the current video surface */
	retval = SDL_MapSurface(SDL_ShadowSurface, SDL_VideoSurface->format);

	/* Save and reset the mapping */
	SDL_FreeSurfaceMap(&shadow_map);
	SDL_SaveSurfaceMap(&shadow_map, SDL_ShadowSurface);
	SDL_RestoreSurfaceMap(&user_map, SDL_ShadowSurface);

	/* Done! */
	return(retval);
}

/*
 * Create a shadow surface suitable for fooling the app. :-)
 */
static void SDL_AllocShadowSurface(Uint8 depth)
{
	Uint32 Rmask, Gmask, Bmask;

	/* Allocate the shadow surface */
	if ( depth == (SDL_VideoSurface->format)->BitsPerPixel ) {
		Rmask = (SDL_VideoSurface->format)->Rmask;
		Gmask = (SDL_VideoSurface->format)->Gmask;
		Bmask = (SDL_VideoSurface->format)->Bmask;
	} else {
		Rmask = Gmask = Bmask = 0;
	}
	SDL_ShadowSurface = SDL_AllocSurface(SDL_SWSURFACE,
				SDL_VideoSurface->w, SDL_VideoSurface->h,
						depth, Rmask, Gmask, Bmask, 0);
	if ( SDL_ShadowSurface == NULL ) {
		return;
	}

	/* 8-bit shadow surfaces report that they have exclusive palette */
	if ( depth == 8 ) {
		SDL_ShadowSurface->flags |= SDL_HWPALETTE;
		if ( depth == (SDL_VideoSurface->format)->BitsPerPixel ) {
			memcpy(SDL_ShadowSurface->format->palette->colors,
				SDL_VideoSurface->format->palette->colors,
				SDL_VideoSurface->format->palette->ncolors*
							sizeof(SDL_Color));
		} else {
			SDL_DitherColors(
			SDL_ShadowSurface->format->palette->colors, depth);
		}
	}

	/* If the video surface is fullscreen, the shadow should say so */
	if ( (SDL_VideoSurface->flags & SDL_FULLSCREEN) == SDL_FULLSCREEN ) {
		SDL_ShadowSurface->flags |= SDL_FULLSCREEN;
	}
	SDL_ClearSurface(SDL_ShadowSurface);
	
	/* Map the shadow surface to the display surface format */
	SDL_EmptySurfaceMap(&shadow_map);
	SDL_MapShadowSurface();
	return;
}

static void SDL_FreeShadowSurface(void)
{
	if ( SDL_ShadowSurface ) {
		SDL_FreeSurface(SDL_ShadowSurface);
		SDL_FreeSurfaceMap(&shadow_map);
		SDL_ShadowSurface = NULL;
	}
}

/*
 * Update a specific portion of the physical screen
 */
void SDL_UpdateRect(SDL_Surface *screen, Sint32 x, Sint32 y, Uint32 w, Uint32 h)
{
	if ( screen ) {
		SDL_Rect rect;

		/* Perform some checking */
		if ( w == 0 )
			w = screen->w;
		if ( h == 0 )
			h = screen->h;
		if ( (x+w) > screen->w )
			return;
		if ( (y+h) > screen->h )
			return;

		/* Fill the rectangle */
		rect.x = x;
		rect.y = y;
		rect.w = w;
		rect.h = h;
		SDL_UpdateRects(screen, 1, &rect);
	}
}
void SDL_UpdateRects (SDL_Surface *screen, int numrects, SDL_Rect *rects)
{
	int i;

	if ( screen == SDL_ShadowSurface ) {
		/* Replace mapping with our own for this blit */
		SDL_SaveSurfaceMap(&user_map, SDL_ShadowSurface);
		SDL_RestoreSurfaceMap(&shadow_map, SDL_ShadowSurface);
		
		for ( i=0; i<numrects; ++i ) {
			SDL_MiddleBlit(SDL_ShadowSurface, &rects[i], 
						SDL_VideoSurface, &rects[i]);
		}

		/* Replace the user's shadow surface mapping */
		SDL_RestoreSurfaceMap(&user_map, SDL_ShadowSurface);

		/* Fall through to video surface update */
		screen = SDL_VideoSurface;
	}
	if ( screen == SDL_VideoSurface ) {
		/* Update the video surface */
		if ( screen->offset ) {
			for ( i=0; i<numrects; ++i ) {
				rects[i].x += offset_x;
				rects[i].y += offset_y;
			}
			(*SDL_SYS_UpdateRects)(screen, numrects, rects);
			for ( i=0; i<numrects; ++i ) {
				rects[i].x -= offset_x;
				rects[i].y -= offset_y;
			}
		} else {
			(*SDL_SYS_UpdateRects)(screen, numrects, rects);
		}
	}
}

/*
 * Set the display surface colormap.
 */
int SDL_SetColors (SDL_Surface *screen,
				SDL_Color *colors, int firstcolor, int ncolors)
{
	int set_video;
	int alloct_all;

	/* Fail if there's no palette to set */
	if ( screen->format->palette == NULL )
		return(0);

	/* Trying to set too many colors? */
	alloct_all = 1;
	if ( (firstcolor+ncolors) > screen->format->palette->ncolors ) {
		/* My mind tells me no!
		   But my heart, and my hips, cry proceed!
			-- The Simpsons
		 */
		ncolors = screen->format->palette->ncolors - firstcolor;
		alloct_all = 0;
	}

	/* Copy the colors to the surface */
	memcpy(&screen->format->palette->colors[firstcolor], colors,
						ncolors*sizeof(*colors));

	/* Check to see if we need to update the display colormap */
	if ( screen == SDL_ShadowSurface ) {
		/* If display isn't an identity palette, remap pixels */
		set_video = SDL_SetColors(SDL_VideoSurface, colors,
						firstcolor, ncolors);
		if ( ! set_video ) {
			SDL_MapShadowSurface();
			SDL_UpdateRect(screen, 0, 0, 0, 0);
		}
	} else
	if ( screen == SDL_VideoSurface ) {
		set_video = SDL_SYS_SetColors(SDL_VideoSurface,
						firstcolor, ncolors);
		if ( ! set_video ) {
			alloct_all = 0;
		}
		SDL_RecolorCursor();
	}

	/* We're done! */
	return(alloct_all);
}

/*
 * Clean up the video subsystem
 */
void SDL_VideoQuit (void)
{
	/* Halt event processing before doing anything else */
	SDL_StopEventLoop();

	/* Clean up allocated window manager items */
	if ( SDL_PublicSurface ) {
		SDL_PublicSurface = NULL;
	}
	SDL_CursorQuit();

	/* Clean up the system video */
	SDL_SYS_VideoQuit(SDL_VideoSurface);

	/* Free any lingering surfaces */
	if ( SDL_VideoSurface != NULL ) {
		SDL_FreeSurface(SDL_VideoSurface);
		SDL_VideoSurface = NULL;
	}
	SDL_FreeShadowSurface();
	if ( SDL_FadedSurface != NULL ) {
		SDL_FreeSurface(SDL_FadedSurface);
		SDL_FadedSurface = NULL;
	}
	SDL_PublicSurface = NULL;

	/* Clean up miscellaneous memory */
	if ( wm_title != NULL ) {
		free(wm_title);
	}
	if ( wm_icon != NULL ) {
		free(wm_icon);
	}

	/* Finish cleaning up video subsystem */
	SDL_SYS_FinalQuit();

	/* That's all folks */
	return;
}

/*
 * Sets/Gets the title and icon text of the display window, if any.
 */
void SDL_WM_SetCaption (const char *title, const char *icon)
{
	if ( title ) {
		if ( wm_title ) {
			free(wm_title);
		}
		wm_title = (char *)malloc(strlen(title)+1);
		if ( wm_title != NULL ) {
			strcpy(wm_title, title);
		}
	}
	if ( icon ) {
		if ( wm_icon ) {
			free(wm_icon);
		}
		wm_icon = (char *)malloc(strlen(icon)+1);
		if ( wm_icon != NULL ) {
			strcpy(wm_icon, icon);
		}
	}
	if ( (title || icon) && (SDL_WMCaps.SetCaption != NULL) ) {
		(*SDL_WMCaps.SetCaption)(SDL_VideoSurface, wm_title, wm_icon);
	}
}
void SDL_WM_GetCaption (char **title, char **icon)
{
	if ( title ) {
		*title = wm_title;
	}
	if ( icon ) {
		*icon = wm_icon;
	}
}

/*
 * Sets the window manager icon for the display window.
 */
void SDL_WM_SetIcon (SDL_Surface *icon, Uint8 *mask)
{
	if ( SDL_WMCaps.SetIcon != NULL ) {
		/* Generate a mask if necessary, and create the icon! */
		if ( mask == NULL ) {
			int mask_len = icon->h*(icon->w+7)/8;
			mask = (Uint8 *)malloc(mask_len);
			if ( mask == NULL ) {
				return;
			}
			memset(mask, ~0, mask_len);
			(*SDL_WMCaps.SetIcon)(SDL_VideoSurface, icon, mask);
			free(mask);
		} else {
			(*SDL_WMCaps.SetIcon)(SDL_VideoSurface, icon, mask);
		}
	}
}

/*
 * Get some platform dependent window manager information
 */
int SDL_GetWMInfo (SDL_SysWMinfo *info)
{
	if ( SDL_WMCaps.GetWMInfo != NULL ) {
		return((*SDL_WMCaps.GetWMInfo)(info));
	} else {
		return(0);
	}
}

/*
 * These functions are placeholders for video drivers that don't have
 * access to the hardware.
 */
static int SDL_NoHWAlloc(SDL_Surface *surface)
{
	return(-1);
}
static int SDL_NoHWAccel(SDL_Surface *surface)
{
	return(0);
}
