/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_wave.c,v 1.10 1998/08/18 15:01:40 slouken Exp $";
#endif

/* Microsoft WAVE file loading routines */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SDL_error.h"
#include "SDL_audio.h"
#include "SDL_wave.h"
#include "SDL_endian.h"

static int ReadChunk(FILE *fp, Chunk *chunk);

struct ADPCM_decodestate {
	Uint8 hPredictor;
	Uint16 iDelta;
	Sint16 iSamp1;
	Sint16 iSamp2;
};
static struct ADPCM_decoder {
	WaveFMT wavefmt;
	Uint16 wSamplesPerBlock;
	Uint16 wNumCoef;
	Sint16 aCoeff[7][2];
	/* * * */
	struct ADPCM_decodestate state[2];
} ADPCM_state;

static int InitADPCM(WaveFMT *format)
{
	Uint8 *rogue_feel;
	Uint16 extra_info;
	int i;

	/* Set the rogue pointer to the ADPCM specific data */
	ADPCM_state.wavefmt.encoding = SDL_SwapLE16(format->encoding);
	ADPCM_state.wavefmt.channels = SDL_SwapLE16(format->channels);
	ADPCM_state.wavefmt.frequency = SDL_SwapLE32(format->frequency);
	ADPCM_state.wavefmt.byterate = SDL_SwapLE32(format->byterate);
	ADPCM_state.wavefmt.blockalign = SDL_SwapLE16(format->blockalign);
	ADPCM_state.wavefmt.bitspersample = SDL_SwapLE16(format->blockalign);
	rogue_feel = (Uint8 *)format+sizeof(*format);
	if ( sizeof(*format) == 16 ) {
		extra_info = ((rogue_feel[1]<<8)|rogue_feel[0]);
		rogue_feel += sizeof(Uint16);
	}
	ADPCM_state.wSamplesPerBlock = ((rogue_feel[1]<<8)|rogue_feel[0]);
	rogue_feel += sizeof(Uint16);
	ADPCM_state.wNumCoef = ((rogue_feel[1]<<8)|rogue_feel[0]);
	rogue_feel += sizeof(Uint16);
	if ( ADPCM_state.wNumCoef != 7 ) {
		SDL_SetError("Unknown set of ADPCM coefficients");
		return(-1);
	}
	for ( i=0; i<ADPCM_state.wNumCoef; ++i ) {
		ADPCM_state.aCoeff[i][0] = ((rogue_feel[1]<<8)|rogue_feel[0]);
		rogue_feel += sizeof(Uint16);
		ADPCM_state.aCoeff[i][1] = ((rogue_feel[1]<<8)|rogue_feel[0]);
		rogue_feel += sizeof(Uint16);
	}
	return(0);
}

static Sint32 ADPCM_nibble(struct ADPCM_decodestate *state,
					Sint8 nybble, Sint16 *coeff)
{
	const Sint32 max_audioval = ((1<<(16-1))-1);
	const Sint32 min_audioval = -(1<<(16-1));
	const Sint32 adaptive[] = {
		230, 230, 230, 230, 307, 409, 512, 614,
		768, 614, 512, 409, 307, 230, 230, 230
	};
	Sint32 new_sample, delta;

	new_sample = ((state->iSamp1 * coeff[0]) +
		      (state->iSamp2 * coeff[1]))/256;
	if ( nybble & 0x08 ) {
		new_sample += state->iDelta * (nybble-0x10);
	} else {
		new_sample += state->iDelta * nybble;
	}
	if ( new_sample < min_audioval ) {
		new_sample = min_audioval;
	} else
	if ( new_sample > max_audioval ) {
		new_sample = max_audioval;
	}
	delta = ((Sint32)state->iDelta * adaptive[nybble])/256;
	if ( delta < 16 ) {
		delta = 16;
	}
	state->iDelta = delta;
	state->iSamp2 = state->iSamp1;
	state->iSamp1 = new_sample;
	return(new_sample);
}

static int ADPCM_decode(Uint8 **audio_buf, Uint32 *audio_len)
{
	struct ADPCM_decodestate *state[2];
	Uint8 *encoded, *decoded;
	Sint32 encoded_len, samplesleft;
	Sint8 nybble, stereo;
	Sint16 *coeff[2];
	Sint32 new_sample;

	/* Allocate the proper sized output buffer */
	encoded = *audio_buf;
	encoded_len = *audio_len;
	*audio_len = (encoded_len/ADPCM_state.wavefmt.blockalign) * 
				ADPCM_state.wSamplesPerBlock*
				ADPCM_state.wavefmt.channels*sizeof(Sint16);
	*audio_buf = (Uint8 *)malloc(*audio_len);
	if ( *audio_buf == NULL ) {
		SDL_OutOfMemory();
		return(-1);
	}
	decoded = *audio_buf;

	/* Get ready... Go! */
	stereo = (ADPCM_state.wavefmt.channels == 2);
	state[0] = &ADPCM_state.state[0];
	state[1] = &ADPCM_state.state[stereo];
	while ( encoded_len >= ADPCM_state.wavefmt.blockalign ) {
		/* Grab the initial information for this block */
		state[0]->hPredictor = *encoded++;
		if ( stereo ) {
			state[1]->hPredictor = *encoded++;
		}
		state[0]->iDelta = ((encoded[1]<<8)|encoded[0]);
		encoded += sizeof(Sint16);
		if ( stereo ) {
			state[1]->iDelta = ((encoded[1]<<8)|encoded[0]);
			encoded += sizeof(Sint16);
		}
		state[0]->iSamp1 = ((encoded[1]<<8)|encoded[0]);
		encoded += sizeof(Sint16);
		if ( stereo ) {
			state[1]->iSamp1 = ((encoded[1]<<8)|encoded[0]);
			encoded += sizeof(Sint16);
		}
		state[0]->iSamp2 = ((encoded[1]<<8)|encoded[0]);
		encoded += sizeof(Sint16);
		if ( stereo ) {
			state[1]->iSamp2 = ((encoded[1]<<8)|encoded[0]);
			encoded += sizeof(Sint16);
		}
		coeff[0] = ADPCM_state.aCoeff[state[0]->hPredictor];
		coeff[1] = ADPCM_state.aCoeff[state[1]->hPredictor];

		/* Store the two initial samples we start with */
		decoded[0] = state[0]->iSamp2&0xFF;
		decoded[1] = state[0]->iSamp2>>8;
		decoded += 2;
		if ( stereo ) {
			decoded[0] = state[1]->iSamp2&0xFF;
			decoded[1] = state[1]->iSamp2>>8;
			decoded += 2;
		}
		decoded[0] = state[0]->iSamp1&0xFF;
		decoded[1] = state[0]->iSamp1>>8;
		decoded += 2;
		if ( stereo ) {
			decoded[0] = state[1]->iSamp1&0xFF;
			decoded[1] = state[1]->iSamp1>>8;
			decoded += 2;
		}

		/* Decode and store the other samples in this block */
		samplesleft = (ADPCM_state.wSamplesPerBlock-2)*
					ADPCM_state.wavefmt.channels;
		while ( samplesleft > 0 ) {
			nybble = (*encoded)>>4;
			new_sample = ADPCM_nibble(state[0], nybble, coeff[0]);
			decoded[0] = new_sample&0xFF;
			new_sample >>= 8;
			decoded[1] = new_sample&0xFF;
			decoded += 2;

			nybble = (*encoded)&0x0F;
			new_sample = ADPCM_nibble(state[1], nybble, coeff[1]);
			decoded[0] = new_sample&0xFF;
			new_sample >>= 8;
			decoded[1] = new_sample&0xFF;
			decoded += 2;

			++encoded;
			samplesleft -= 2;
		}
		encoded_len -= ADPCM_state.wavefmt.blockalign;
	}
	return(0);
}

SDL_AudioSpec * SDL_LoadWAV (const char *file,
		SDL_AudioSpec *spec, Uint8 **audio_buf, Uint32 *audio_len)
{
	FILE *wave;
	Chunk chunk;
	int lenread;
	int ADPCM_encoded;

	/* WAV magic header */
	Uint32 RIFFchunk;
	Uint32 wavelen;
	Uint32 WAVEmagic;

	/* FMT chunk */
	WaveFMT *format;

	/* Open the WAVE file for reading */
	wave = fopen(file, "rb");
	if ( wave == NULL ) {
		SDL_SetError("Couldn't open for reading");
		return(NULL);
	}
	
	/* Check the magic header */
	SDL_CalculateEndian();
	RIFFchunk	= SDL_ReadLE32(wave);
	wavelen		= SDL_ReadLE32(wave);
	WAVEmagic	= SDL_ReadLE32(wave);
	if ( (RIFFchunk != RIFF) || (WAVEmagic != WAVE) ) {
		SDL_SetError("Unrecognized file type (not WAVE)");
		fclose(wave);
		return(NULL);
	}

	/* Read the audio data format chunk */
	do {
		lenread = ReadChunk(wave, &chunk);
		if ( lenread < 0 ) {
			fclose(wave);
			return(NULL);
		}
	} while ( chunk.magic == FACT );

	/* Decode the audio data format */
	format = (WaveFMT *)chunk.data;
	if ( chunk.magic != FMT ) {
		SDL_SetError("Complex WAVE files not supported");
		free(format);
		fclose(wave);
		return(NULL);
	}
	ADPCM_encoded = 0;
	switch (SDL_SwapLE16(format->encoding)) {
		case PCM_CODE:
			/* We can understand this */
			break;
		case ADPCM_CODE:
			/* We can try to understand this */
			if ( InitADPCM(format) < 0 ) {
				free(format);
				fclose(wave);
				return(NULL);
			}
			ADPCM_encoded = 1;
			break;
		default:
			SDL_SetError("Unknown WAVE data format");
			free(format);
			fclose(wave);
			return(NULL);
	}
	memset(spec, 0, sizeof(*spec));
	spec->freq = SDL_SwapLE32(format->frequency);
	switch (SDL_SwapLE16(format->bitspersample)) {
		case 4:
			if ( ADPCM_encoded ) {
				spec->format = AUDIO_S16;
			} else {
				SDL_SetError("Unknown PCM data format");
				free(format);
				fclose(wave);
				return(NULL);
			}
			break;
		case 8:
			spec->format = AUDIO_U8;
			break;
		case 16:
			spec->format = AUDIO_S16;
			break;
		default:
			SDL_SetError("Unknown PCM data format");
			free(format);
			fclose(wave);
			return(NULL);
	}
	switch (SDL_SwapLE16(format->channels)) {
		case 1:
			spec->format |= AUDIO_MONO;
			break;
		case 2:
			spec->format |= AUDIO_STEREO;
			break;
		default:
			SDL_SetError("Strange number of audio channels");
			free(format);
			fclose(wave);
			return(NULL);
	}
	spec->samples	= 4096;		/* Good default buffer size */
	free(format);

	/* Read the audio data chunk */
	*audio_buf = NULL;
	do {
		if ( *audio_buf != NULL ) {
			free(*audio_buf);
		}
		lenread = ReadChunk(wave, &chunk);
		if ( lenread < 0 ) {
			fclose(wave);
			return(NULL);
		}
		*audio_len = lenread;
		*audio_buf = chunk.data;
	} while ( chunk.magic != DATA );

	fclose(wave);

	if ( ADPCM_encoded ) {
		if ( ADPCM_decode(audio_buf, audio_len) < 0 ) {
			return(NULL);
		}
	}
	return(spec);
}

static int ReadChunk(FILE *fp, Chunk *chunk)
{
	chunk->magic	= SDL_ReadLE32(fp);
	chunk->length	= SDL_ReadLE32(fp);
	chunk->data = (Uint8 *)malloc(chunk->length);
	if ( chunk->data == NULL ) {
		SDL_OutOfMemory();
		return(-1);
	}
	if ( fread(chunk->data, chunk->length, 1, fp) != 1 ) {
		SDL_SetError("Error reading chunk -- truncated file?");
		free(chunk->data);
		return(-1);
	}
	return(chunk->length);
}
