/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_thread.c,v 1.3 1998/08/18 15:33:59 slouken Exp $";
#endif

/* BeOS thread management routines for SDL */

#include <stdio.h>
#include <signal.h>
#include <be/kernel/OS.h>

#include "SDL_error.h"
#include "SDL_mutex.h"
#include "SDL_thread.h"


int sig_list[] = {
	SIGHUP, SIGINT, SIGQUIT, SIGPIPE, SIGALRM, SIGTERM, SIGWINCH, 0
};

struct SDL_Thread {
	thread_id id;
};

typedef struct {
	void (*func)(void *);
	void  *data;
	SDL_mutex *wait;
} thread_args;

static int32 RunThread(void *data)
{
	thread_args *args;
	void (*userfunc)(void *);
	void  *userdata;
	int i;
	sigset_t mask;

	/* Figure out what function to run */
	args = (thread_args *)data;
	userfunc = args->func;
	userdata = args->data;

	/* Mask asynchronous signals for this thread */
	sigemptyset(&mask);
	for ( i=0; sig_list[i]; ++i ) {
		sigaddset(&mask, sig_list[i]);
	}
	sigprocmask(SIG_BLOCK, &mask, NULL);

	/* Release the parent process */
	SDL_mutexV(args->wait);

	/* Run the function */
	(*userfunc)(userdata);
	exit_thread(0);
}

SDL_Thread *SDL_CreateThread(void (*fn)(void *), void *data)
{
	SDL_Thread *thread;
	thread_args *args;

	/* Allocate memory for the thread info structure */
	thread = (SDL_Thread *)malloc(sizeof(*thread));
	if ( thread == NULL )
		return(NULL);

	/* Set up the arguments for the thread */
	args = (thread_args *)malloc(sizeof(*args));
	if ( args == NULL ) {
		free(thread);
		return(NULL);
	}
	args->func = fn;
	args->data = data;
	args->wait = SDL_CreateMutex();
	if ( args->wait == NULL ) {
		free(thread);
		free(args);
		return(NULL);
	}
	SDL_mutexP(args->wait);		/* We will block later */

	/* Create the thread */
	thread->id = spawn_thread(RunThread, "SDL", B_NORMAL_PRIORITY, args);
	if ((thread->id == B_NO_MORE_THREADS) || (thread->id == B_NO_MEMORY)) {
		SDL_SetError("Out of resources");
		free(thread);
		thread = NULL;
	} else {
		resume_thread(thread->id);
		SDL_mutexP(args->wait);
	}
	SDL_DestroyMutex(args->wait);
	free(args);

	/* Everything is running now */
	return(thread);
}

void SDL_WaitThread(SDL_Thread *thread)
{
	status_t status;

	if ( thread ) {
		wait_for_thread(thread->id, &status);
		free(thread);
	}
}

void SDL_KillThread(SDL_Thread *thread)
{
	if ( thread ) {
		kill_thread(thread->id);
		SDL_WaitThread(thread);
	}
}
