/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_sysvideo.cc,v 1.1.1.1 1998/07/14 01:51:17 slouken Exp $";
#endif

/* BWindow based framebuffer implementation */

#include <stdlib.h>
#include <string.h>

#include <stdio.h>
#include <unistd.h>

#include "SDL.h"
#include "SDL_BWin.h"
#include "SDL_thread.h"
#include "SDL_timer.h"
#include "blank_cursor.h"
#include "SDL_lowvideo.h"

extern "C" {

#include "SDL_sysvideo.h"
#include "SDL_sysmouse_c.h"
#include "SDL_syswm_c.h"

#define BEOS_HIDDEN_SIZE	32	/* starting hidden window size */

/* Local data members -- actually covers the entire window interaction */
static SDL_Thread *SDL_AppThread = NULL;
SDL_BWin   *SDL_Win = NULL;
static BBitmap *SDL_Bitmap = NULL;
WMcursor *SDL_BlankCursor = NULL;

static void StartApp(void *unused)
{
	BApplication *App;

	App = new BApplication("application/x-SDL-executable");
	App->Run();
	delete App;
}

int SDL_SYS_VideoInit(SDL_PixelFormat *vformat)
{
	void SDL_VideoQuit(void);
	BRect bounds;

	/* Create the BApplication that handles appserver interaction */
	if ( be_app != NULL ) { /* Not tested */
		be_app->PostMessage(B_QUIT_REQUESTED);
		while ( be_app != NULL ) {
			SDL_Delay(20);
		}
	}
	SDL_AppThread = SDL_CreateThread(StartApp, NULL);
	if ( SDL_AppThread == NULL ) {
		SDL_SetError("Couldn't create BApplication thread");
	}
	do {
		SDL_Delay(10);
	} while ( (be_app == NULL) || be_app->IsLaunching() );

	/* Determine the screen depth */
	{
	  BScreen bscreen;
	  switch(bscreen.ColorSpace()) {
		case B_CMAP8:
			vformat->BitsPerPixel = 8;
			break;
		case B_RGB15:
		case B_RGBA15:
		case B_RGB15_BIG:
		case B_RGBA15_BIG:
			vformat->BitsPerPixel = 15;
			break;
		case B_RGB16:
		case B_RGB16_BIG:
			vformat->BitsPerPixel = 16;
			break;
		case B_RGB32:
		case B_RGBA32:
		case B_RGB32_BIG:
		case B_RGBA32_BIG:
			vformat->BitsPerPixel = 32;
			break;
		default:
			SDL_SetError("Unknown BScreen colorspace: 0x%x",
							bscreen.ColorSpace());
			return(-1);
	  }
	}

	/* Create the window and view */
	bounds.top = 0; bounds.left = 0;
	bounds.right = BEOS_HIDDEN_SIZE;
	bounds.bottom = BEOS_HIDDEN_SIZE;
	SDL_Win = new SDL_BWin(bounds);

	/* Create the clear cursor */
	SDL_BlankCursor = SDL_CreateWMCursor(NULL, blank_cdata, blank_cmask,
			BLANK_CWIDTH, BLANK_CHEIGHT, BLANK_CHOTX, BLANK_CHOTY);

	/* Fill in some window manager capabilities */
	SDL_WMCaps.SetCaption = SDL_SYS_SetWMCaption;

	/* We're done! */
	return(0);
}

/* We support any dimension at our bit-depth */
SDL_Rect **SDL_SYS_ListModes(SDL_Surface *screen,
					SDL_PixelFormat *format, Uint32 flags)
{
	if ( format->BitsPerPixel == screen->format->BitsPerPixel ) {
		return((SDL_Rect **)-1);
	} else {
		return((SDL_Rect **)0);
	}
}

/* Various screen update functions available */
static void SDL_NormalUpdate(SDL_Surface *screen,int numrects,SDL_Rect *rects);

/* FIXME: check return values and cleanup here */
SDL_Surface *SDL_SYS_SetVideoMode(SDL_Surface *current,
			Uint16 width, Uint16 height, Uint8 bpp, Uint32 flags)
{
	BScreen bscreen;
	BRect bounds;

	/* Create the BBitmap framebuffer */
	if ( SDL_Bitmap != NULL ) {
		delete SDL_Bitmap;
	}
	bounds.top = 0; bounds.left = 0;
	bounds.right = width-1;
	bounds.bottom = height-1;
	SDL_Bitmap = new BBitmap(bounds, bscreen.ColorSpace());
	if ( ! SDL_Bitmap->IsValid() ) {
		SDL_SetError("Couldn't create screen bitmap");
		delete SDL_Bitmap;
		return(NULL);
	}
	current->flags &= SDL_THREADSAFE;	/* Clear flags */
	current->w = width;
	current->h = height;
	current->pitch = SDL_Bitmap->BytesPerRow();
	current->pixels = (void *)SDL_Bitmap->Bits();
	SDL_Win->SetBitmap(SDL_Bitmap);

	/* Determine what blit function to use */
	SDL_SYS_UpdateRects = SDL_NormalUpdate;

	/* Hide the window for resizing */
	if ( SDL_Win->Lock() ) {
		if ( ! SDL_Win->IsHidden() )
			SDL_Win->Hide();

		/* Resize and show! */
		SDL_Win->ResizeTo(width, height);
		bounds = bscreen.Frame();
		SDL_Win->MoveTo((bounds.Width()-width)/2,
				(bounds.Height()-height)/2);
		SDL_Win->Show();
		if ( SDL_Win->CountLocks() > 0 ) {
			SDL_Win->Unlock();
		}
	}


	/* We're done */
	return(current);
}

static void SDL_NormalUpdate(SDL_Surface *screen,int numrects,SDL_Rect *rects)
{
	if ( SDL_BWin::Alive() && SDL_Win->BeginDraw() ) {
		int i;

		for ( i=0; i<numrects; ++i ) {
			BRect rect;

			rect.top = rects[i].y;
			rect.left = rects[i].x;
			rect.bottom = rect.top+rects[i].h-1;
			rect.right = rect.left+rects[i].w-1;
			SDL_Win->DrawAsync(rect);
		}
		SDL_Win->EndDraw();
	}
}

/* Is the system palette settable? */
int SDL_SYS_SetColors(SDL_Surface *screen, int firstcolor, int ncolors)
{
	int i;
	SDL_Palette *palette;
	const color_map *cmap = BScreen().ColorMap();

	/* Get the screen colormap */
	palette = screen->format->palette;
	for ( i=0; i<256; ++i ) {
		palette->colors[i].r = cmap->color_list[i].red;
		palette->colors[i].g = cmap->color_list[i].green;
		palette->colors[i].b = cmap->color_list[i].blue;
	}
	return(0);
}

void SDL_SYS_VideoQuit(SDL_Surface *screen)
{
	if ( SDL_BlankCursor != NULL ) {
		SDL_FreeWMCursor(screen, SDL_BlankCursor);
		SDL_BlankCursor = NULL;
	}
	if ( screen ) {
		screen->pixels = NULL;
	}
}
void SDL_SYS_FinalQuit(void) { }


}; /* Extern C */
