/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_sysevents.c,v 1.5 1998/09/11 12:13:40 slouken Exp $";
#endif

/* Handle the event stream, converting X11 events into SDL events */

#include <stdio.h>
#include <setjmp.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#ifdef __SVR4
#include <X11/Sunkeysym.h>
#endif
#include <sys/time.h>

#include "SDL.h"
#include "SDL_sysevents.h"
#include "SDL_video_c.h"
#include "SDL_events_c.h"
#include "SDL_syswm.h"
#include "SDL_lowvideo.h"


/* The translation tables from an X11 keysym to a SDL keysym */
static SDLKey ASCII_keymap[128];
static SDLKey MISC_keymap[256];
static SDLKey TransKey(XKeyEvent *xkey);

/* These variables allow us to jump out of a blocking X11 event wait */
static jmp_buf interrupt_buf;
static int x_waiting;
void SDL_SYS_QuitInterrupt(void)
{
	/* There's a slight race condition with SDL_DispatchEvent(), it's ok */
	if ( x_waiting ) {
		longjmp(interrupt_buf, 1);
	}
}

static int SDL_DispatchEvent(void)
{
	int posted;
	XEvent xevent;

	/* Allow ourselves to be interrupted */
	if ( setjmp(interrupt_buf) != 0 ) {
		return(0);
	}
	x_waiting = 1;
	XNextEvent(SDL_Display, &xevent);
	x_waiting = 0;

	posted = 0;
	switch (xevent.type) {

		/* Gaining mouse coverage? */
		case EnterNotify: {
			posted = SDL_PrivateAppActive(1, SDL_APPMOUSEFOCUS);
		}
		break;

		/* Gaining input focus? */
		case FocusIn: {
			/* Check for keyboard auto-repeat, and turn it off */
			if ( ! (SDL_GetAppState() & SDL_APPINPUTFOCUS) ) {
				XKeyboardState kbstate;

				XGetKeyboardControl(SDL_Display, &kbstate);
				if ( kbstate.global_auto_repeat ) {
					X11_key_repeat = 1;
				} else {
					X11_key_repeat = 0;
				}
				XAutoRepeatOff(SDL_Display);
				XSync(SDL_Display, False);
			}
			posted = SDL_PrivateAppActive(1, SDL_APPINPUTFOCUS);
		}
		break;

		/* Losing mouse coverage? */
		case LeaveNotify: {
			posted = SDL_PrivateAppActive(0, SDL_APPMOUSEFOCUS);
		}
		break;

		/* Losing input focus? */
		case FocusOut: {
			if ( X11_key_repeat ) {
				XAutoRepeatOn(SDL_Display);
				XSync(SDL_Display, False);
			}
			posted = SDL_PrivateAppActive(0, SDL_APPINPUTFOCUS);
		}
		break;

		/* Mouse motion? */
		case MotionNotify: {
			if ( SDL_VideoSurface->flags & SDL_FULLSCREEN ) {
				posted = SDL_PrivateMouseMotion(0, 1,
							xevent.xmotion.x_root,
							xevent.xmotion.y_root);
			} else {
				posted = SDL_PrivateMouseMotion(0, 0,
							xevent.xmotion.x,
							xevent.xmotion.y);
			}
		}
		break;

		/* Mouse button press? */
		case ButtonPress: {
			posted = SDL_PrivateMouseButton(SDL_PRESSED, 
						xevent.xbutton.button, 0, 0);
		}
		break;

		/* Mouse button release? */
		case ButtonRelease: {
			posted = SDL_PrivateMouseButton(SDL_RELEASED, 
						xevent.xbutton.button, 0, 0);
		}
		break;

		/* Key press? */
		case KeyPress: {
			posted = SDL_PrivateKeyboard(SDL_PRESSED, 
						TransKey(&xevent.xkey));
		}
		break;

		/* Key release? */
		case KeyRelease: {
			posted = SDL_PrivateKeyboard(SDL_RELEASED, 
						TransKey(&xevent.xkey));
		}
		break;

		/* Have we been iconified? */
		case UnmapNotify: {
			posted = SDL_PrivateAppActive(0,
					(SDL_APPACTIVE|SDL_APPMOUSEFOCUS));
		}
		break;

		/* Have we been restored? */
		case MapNotify: {
			posted = SDL_PrivateAppActive(1,
					(SDL_APPACTIVE|SDL_APPMOUSEFOCUS));
		}
		break;

		/* Have we been requested to quit? */
		case ClientMessage: {
			if ( (xevent.xclient.format == 32) &&
			     (xevent.xclient.data.l[0] == SDL_DELWIN) )
			{
				posted = SDL_PrivateQuit();
			}
		}
		break;

		/* Do we need to refresh ourselves? */
		case Expose: {
			if ( xevent.xexpose.count == 0 ) {
				SDL_UpdateRect(SDL_PublicSurface, 0, 0, 0, 0);
			}
		}
		break;

		default: {
			/* Only post the event if we're watching for it */
			if ( SDL_ProcessEvents[SDL_SYSWMEVENT] == SDL_ENABLE ) {
				SDL_SysWMmsg wmmsg;

				SDL_VERSION(&wmmsg.version);
				wmmsg.xevent = xevent;
				posted = SDL_PrivateSysWMEvent(&wmmsg);
			}
		}
		break;
	}
	return(posted);
}

/* Ack!  XPending() actually performs a blocking read if no events available */
static int SDL_X11_Pending(void)
{
	/* Flush the display connection and look to see if events are queued */
	XFlush(SDL_Display);
	if ( XEventsQueued(SDL_Display, QueuedAlready) ) {
		return(1);
	}

	/* More drastic measures are required -- see if X is ready to talk */
	{
		static struct timeval zero_time;	/* static == 0 */
		int x11_fd;
		fd_set fdset;

		x11_fd = ConnectionNumber(SDL_Display);
		FD_ZERO(&fdset);
		FD_SET(x11_fd, &fdset);
		if ( select(x11_fd+1, &fdset, NULL, NULL, &zero_time) == 1 ) {
			return(XPending(SDL_Display));
		}
	}

	/* Oh well, nothing is ready .. */
	return(0);
}

void SDL_SYS_PollEvent(void)
{
	/* Keep processing pending events until one is posted */
	while ( SDL_X11_Pending() && ! SDL_DispatchEvent() ) {
		/* Event available, but not posted -- keep looking */;
	}
}

int SDL_SYS_WaitEvent(void)
{
	/* Wait for and process X11 events until one is posted */
	while ( ! SDL_DispatchEvent() ) {
		if ( SDL_CheckQuit(NULL) )
			return(0);
	}
	return(1);
}

void SDL_InitOSKeymap(void)
{
	int i;

	/* Map the ASCII keysyms */
	for ( i=0; i<SDL_TABLESIZE(ASCII_keymap, SDLKey); ++i )
		ASCII_keymap[i] = 0;

	ASCII_keymap[XK_space&0xFF] = SDLK_SPACE;
	ASCII_keymap[XK_comma&0xFF] = SDLK_COMMA;
	ASCII_keymap[XK_minus&0xFF] = SDLK_MINUS;
	ASCII_keymap[XK_period&0xFF] = SDLK_PERIOD;
	ASCII_keymap[XK_slash&0xFF] = SDLK_SLASH;
	ASCII_keymap[XK_quoteright&0xFF] = SDLK_QUOTE;
	ASCII_keymap[XK_0&0xFF] = SDLK_0;
	ASCII_keymap[XK_1&0xFF] = SDLK_1;
	ASCII_keymap[XK_2&0xFF] = SDLK_2;
	ASCII_keymap[XK_3&0xFF] = SDLK_3;
	ASCII_keymap[XK_4&0xFF] = SDLK_4;
	ASCII_keymap[XK_5&0xFF] = SDLK_5;
	ASCII_keymap[XK_6&0xFF] = SDLK_6;
	ASCII_keymap[XK_7&0xFF] = SDLK_7;
	ASCII_keymap[XK_8&0xFF] = SDLK_8;
	ASCII_keymap[XK_9&0xFF] = SDLK_9;
	ASCII_keymap[XK_semicolon&0xFF] = SDLK_SEMICOLON;
	ASCII_keymap[XK_equal&0xFF] = SDLK_EQUALS;
	ASCII_keymap[XK_bracketleft&0xFF] = SDLK_LEFTBRACKET;
	ASCII_keymap[XK_backslash&0xFF] = SDLK_BACKSLASH;
	ASCII_keymap[XK_bracketright&0xFF] = SDLK_RIGHTBRACKET;
	ASCII_keymap[XK_grave&0xFF] = SDLK_BACKQUOTE;
	ASCII_keymap[XK_a&0xFF] = SDLK_a;
	ASCII_keymap[XK_b&0xFF] = SDLK_b;
	ASCII_keymap[XK_c&0xFF] = SDLK_c;
	ASCII_keymap[XK_d&0xFF] = SDLK_d;
	ASCII_keymap[XK_e&0xFF] = SDLK_e;
	ASCII_keymap[XK_f&0xFF] = SDLK_f;
	ASCII_keymap[XK_g&0xFF] = SDLK_g;
	ASCII_keymap[XK_h&0xFF] = SDLK_h;
	ASCII_keymap[XK_i&0xFF] = SDLK_i;
	ASCII_keymap[XK_j&0xFF] = SDLK_j;
	ASCII_keymap[XK_k&0xFF] = SDLK_k;
	ASCII_keymap[XK_l&0xFF] = SDLK_l;
	ASCII_keymap[XK_m&0xFF] = SDLK_m;
	ASCII_keymap[XK_n&0xFF] = SDLK_n;
	ASCII_keymap[XK_o&0xFF] = SDLK_o;
	ASCII_keymap[XK_p&0xFF] = SDLK_p;
	ASCII_keymap[XK_q&0xFF] = SDLK_q;
	ASCII_keymap[XK_r&0xFF] = SDLK_r;
	ASCII_keymap[XK_s&0xFF] = SDLK_s;
	ASCII_keymap[XK_t&0xFF] = SDLK_t;
	ASCII_keymap[XK_u&0xFF] = SDLK_u;
	ASCII_keymap[XK_v&0xFF] = SDLK_v;
	ASCII_keymap[XK_w&0xFF] = SDLK_w;
	ASCII_keymap[XK_x&0xFF] = SDLK_x;
	ASCII_keymap[XK_y&0xFF] = SDLK_y;
	ASCII_keymap[XK_z&0xFF] = SDLK_z;

	/* Map the miscellaneous keys */
	for ( i=0; i<SDL_TABLESIZE(MISC_keymap, SDLKey); ++i )
		MISC_keymap[i] = 0;

	MISC_keymap[XK_BackSpace&0xFF] = SDLK_BACKSPACE;
	MISC_keymap[XK_Tab&0xFF] = SDLK_TAB;
	MISC_keymap[XK_Clear&0xFF] = SDLK_CLEAR;
	MISC_keymap[XK_Return&0xFF] = SDLK_RETURN;
	MISC_keymap[XK_Pause&0xFF] = SDLK_PAUSE;
	MISC_keymap[XK_Scroll_Lock&0xFF] = SDLK_SCROLLOCK;
	MISC_keymap[XK_Sys_Req&0xFF] = SDLK_SYSREQ;
	MISC_keymap[XK_Escape&0xFF] = SDLK_ESCAPE;
	MISC_keymap[XK_Delete&0xFF] = SDLK_DELETE;
	MISC_keymap[XK_Home&0xFF] = SDLK_HOME;
	MISC_keymap[XK_Left&0xFF] = SDLK_LEFT;
	MISC_keymap[XK_Up&0xFF] = SDLK_UP;
	MISC_keymap[XK_Right&0xFF] = SDLK_RIGHT;
	MISC_keymap[XK_Down&0xFF] = SDLK_DOWN;
	MISC_keymap[XK_Page_Up&0xFF] = SDLK_PAGEUP;
	MISC_keymap[XK_Page_Down&0xFF] = SDLK_PAGEDOWN;
	MISC_keymap[XK_End&0xFF] = SDLK_END;
	MISC_keymap[XK_Print&0xFF] = SDLK_PRINT;
	MISC_keymap[XK_Insert&0xFF] = SDLK_INSERT;
	MISC_keymap[XK_Menu&0xFF] = SDLK_MENU;
	MISC_keymap[XK_Help&0xFF] = SDLK_HELP;
	MISC_keymap[XK_Break&0xFF] = SDLK_BREAK;
	MISC_keymap[XK_Num_Lock&0xFF] = SDLK_NUMLOCK;
	MISC_keymap[XK_KP_Enter&0xFF] = SDLK_KP_ENTER;
	MISC_keymap[XK_KP_Home&0xFF] = SDLK_KP7;
	MISC_keymap[XK_KP_Left&0xFF] = SDLK_KP4;
	MISC_keymap[XK_KP_Up&0xFF] = SDLK_KP8;
	MISC_keymap[XK_KP_Right&0xFF] = SDLK_KP6;
	MISC_keymap[XK_KP_Down&0xFF] = SDLK_KP2;
	MISC_keymap[XK_KP_Page_Up&0xFF] = SDLK_KP9;
	MISC_keymap[XK_KP_Page_Down&0xFF] = SDLK_KP3;
	MISC_keymap[XK_KP_End&0xFF] = SDLK_KP1;	
	MISC_keymap[XK_KP_Begin&0xFF] = SDLK_KP5;
	MISC_keymap[XK_KP_Insert&0xFF] = SDLK_KP0;
	MISC_keymap[XK_KP_Delete&0xFF] = SDLK_KP_PERIOD;
	MISC_keymap[XK_KP_Multiply&0xFF] = SDLK_KP_MULTIPLY;
	MISC_keymap[XK_KP_Add&0xFF] = SDLK_KP_PLUS;
	MISC_keymap[XK_KP_Subtract&0xFF] = SDLK_KP_MINUS;
	MISC_keymap[XK_KP_Decimal&0xFF] = SDLK_KP_PERIOD;
	MISC_keymap[XK_KP_Divide&0xFF] = SDLK_KP_DIVIDE;
	MISC_keymap[XK_F1&0xFF] = SDLK_F1;
	MISC_keymap[XK_F2&0xFF] = SDLK_F2;
	MISC_keymap[XK_F3&0xFF] = SDLK_F3;
	MISC_keymap[XK_F4&0xFF] = SDLK_F4;
	MISC_keymap[XK_F5&0xFF] = SDLK_F5;
	MISC_keymap[XK_F6&0xFF] = SDLK_F6;
	MISC_keymap[XK_F7&0xFF] = SDLK_F7;
	MISC_keymap[XK_F8&0xFF] = SDLK_F8;
	MISC_keymap[XK_F9&0xFF] = SDLK_F9;
	MISC_keymap[XK_F10&0xFF] = SDLK_F10;
	MISC_keymap[XK_F11&0xFF] = SDLK_F11;
	MISC_keymap[XK_F12&0xFF] = SDLK_F12;
	MISC_keymap[XK_F13&0xFF] = SDLK_F13;
	MISC_keymap[XK_F14&0xFF] = SDLK_F14;
	MISC_keymap[XK_F15&0xFF] = SDLK_F15;
	MISC_keymap[XK_Shift_L&0xFF] = SDLK_LSHIFT;
	MISC_keymap[XK_Shift_R&0xFF] = SDLK_RSHIFT;
	MISC_keymap[XK_Control_L&0xFF] = SDLK_LCTRL;
	MISC_keymap[XK_Control_R&0xFF] = SDLK_RCTRL;
	MISC_keymap[XK_Caps_Lock&0xFF] = SDLK_CAPSLOCK;
	MISC_keymap[XK_Meta_L&0xFF] = SDLK_LMETA;
	MISC_keymap[XK_Meta_R&0xFF] = SDLK_RMETA;
	MISC_keymap[XK_Alt_L&0xFF] = SDLK_LALT;
	MISC_keymap[XK_Alt_R&0xFF] = SDLK_RALT;
}

static SDLKey TransKey(XKeyEvent *xkey)
{
	KeySym xsym;

	xsym = XLookupKeysym(xkey, 0);
#ifdef DEBUG_KEYS
	fprintf(stderr, "Translating key 0x%.4x (%d)\n", xsym, xkey->keycode);
#endif
	switch (xsym>>8) {
		case 0x1005FF:
#ifdef SunXK_F36
			if ( xsym == SunXK_F36 )
				return(SDLK_F11);
#endif
#ifdef SunXK_F37
			if ( xsym == SunXK_F37 )
				return(SDLK_F12);
#endif
			return(0);
		case 0x00:
			return(ASCII_keymap[xsym&0xFF]);
		case 0xFF:
			return(MISC_keymap[xsym&0xFF]);
		default:
			return(0);
	}
}

