/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_sysaudio.c,v 1.6 1998/10/13 21:31:10 slouken Exp $";
#endif

/* Allow access to a raw mixing buffer */

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <strings.h>
#ifdef __SVR4
#include <sys/audioio.h>
#else
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#endif

#include "SDL_endian.h"
#include "SDL_audio.h"
#include "SDL_audiomem.h"
#include "SDL_audio_c.h"

#ifndef _PATH_DEV_DSP
#define _PATH_DEV_DSP	"/dev/audio"
#endif


/* The file descriptor for the audio device */
static int audio_fd = -1;

/* Raw mixing buffer */
static Uint16 audio_fmt;
static Uint8 *mixbuf;
static int    mixlen;
static int ulaw_only;
static int ulaw_size;
static Uint8 *ulaw_buf;


void SDL_SYS_WaitAudio(void)
{
#ifdef AUDIO_DRAIN
	ioctl(audio_fd, AUDIO_DRAIN, 0);
#else
	fd_set fdset;

	FD_ZERO(&fdset);
	FD_SET(audio_fd, &fdset);
	select(audio_fd+1, NULL, &fdset, NULL, NULL);
#endif
}

void SDL_SYS_PlayAudio(void)
{
	static Uint8 snd2au(int sample);

	/* Write the audio data */
	if ( ulaw_only ) {
		/* Assuming that audio_spec.freq >= 8000 Hz */
		int accum, incr, pos;
		Uint8 *aubuf;

		accum = 0;
		incr  = audio_spec.freq/8;
		aubuf = ulaw_buf;
		switch (audio_fmt & 0xFF) {
			case 8: {
				Uint8 *sndbuf;

				sndbuf = mixbuf;
				for ( pos=0; pos < ulaw_size; ++pos ) {
					*aubuf = snd2au((0x80-*sndbuf)*64);
					accum += incr;
					while ( accum > 0 ) {
						accum -= 1000;
						sndbuf += 1;
					}
					aubuf += 1;
				}
			}
			break;
			case 16: {
				Sint16 *sndbuf;

				sndbuf = (Sint16 *)mixbuf;
				for ( pos=0; pos < ulaw_size; ++pos ) {
					*aubuf = snd2au(*sndbuf/4);
					accum += incr;
					while ( accum > 0 ) {
						accum -= 1000;
						sndbuf += 1;
					}
					aubuf += 1;
				}
			}
			break;
		}
		if ( write(audio_fd, ulaw_buf, ulaw_size) < 0 ) {
			/* Assume fatal error, for now */
			SDL_AudioEnabled = 0;
		}
	} else {
		if ( write(audio_fd, mixbuf, mixlen) < 0 ) {
			/* Assume fatal error, for now */
			SDL_AudioEnabled = 0;
		}
	}
}

Uint8 *SDL_SYS_GetAudioBuf(void)
{
	return(mixbuf);
}

void SDL_SYS_CloseAudio(void)
{
	if ( mixbuf != NULL ) {
		SDL_FreeAudioMem(mixbuf);
		mixbuf = NULL;
	}
	if ( ulaw_buf != NULL ) {
		free(ulaw_buf);
		ulaw_buf = NULL;
	}
	close(audio_fd);
}

int SDL_SYS_OpenAudio(SDL_AudioSpec *spec)
{
	char *audiodev;
#ifdef AUDIO_GETINFO
	audio_info_t ulaw_info;
	audio_info_t play_info;
#endif

	/* Initialize our freeable variables, in case we fail*/
	audio_fd = -1;
	mixbuf = NULL;
	ulaw_buf = NULL;

	/* Determine the audio parameters from the AudioSpec */
	switch ( spec->format & 0xFF ) {

		case 8: { /* Unsigned 8 bit audio data */
			spec->format = AUDIO_U8 | (spec->format&AUDIO_STEREO);
		}
		break;

		case 16: { /* Signed 16 bit audio data */
			int endian;

			endian = SDL_CalculateEndian();
			if ( endian == SDL_BIG_ENDIAN ) {
				spec->format = AUDIO_S16MSB |
						(spec->format&AUDIO_STEREO);
			} else {
				spec->format = AUDIO_S16LSB |
						(spec->format&AUDIO_STEREO);
			}
		}
		break;

		default: {
			SDL_SetError("Unsupported audio format");
			return(-1);
		}
	}
	audio_fmt = spec->format & ~AUDIO_STEREO;

	/* Open the audio device */
	if ( (audiodev=getenv("AUDIODEV")) == NULL ) {
		audiodev = _PATH_DEV_DSP;
	}
	audio_fd = open(audiodev, (O_WRONLY|O_NONBLOCK), 0);
	if ( audio_fd < 0 ) {
		SDL_SetError("Couldn't open %s: %s", audiodev, strerror(errno));
		return(-1);
	}

	/* Supposedly, /dev/audio is opened with U-Law settings */
	ulaw_only = 1;
#ifdef AUDIO_GETINFO
	if ( ioctl(audio_fd, AUDIO_GETINFO, &ulaw_info) < 0 ) {
		SDL_SetError("Couldn't get audio info: %s", strerror(errno));
		return(-1);
	}
	if ( (ulaw_info.play.sample_rate != 8000) ||
	     (ulaw_info.play.channels != 1) ||
	     (ulaw_info.play.precision != 8) ||
	     (ulaw_info.play.encoding != AUDIO_ENCODING_ULAW) ) {
		SDL_SetError("Audio device not in U-Law mode");
		return(-1);
	}
	memcpy(&play_info, &ulaw_info, sizeof(ulaw_info));

	/* Try to set the requested settings */
	play_info.play.sample_rate = spec->freq;
	play_info.play.channels = 1;
	if ( (spec->format&AUDIO_STEREO) == AUDIO_STEREO ) {
		play_info.play.channels *= 2;
	}
	play_info.play.precision = (spec->format&0xFF);
	play_info.play.encoding = AUDIO_ENCODING_LINEAR;
#ifdef AUDIO_DEBUG
  	fprintf(stderr,
	"Trying to set %s %d Hz %d bit %s encoding audio format\n",
	play_info.play.channels == 1 ? "mono" : "stereo",
	play_info.play.sample_rate, play_info.play.precision,
	play_info.play.encoding == AUDIO_ENCODING_LINEAR ? "linear" : "U-law");
#endif
	/* Apparently the dbri audio driver can't do linear 8-bit */
	if ( ((spec->format & 0xFF) != AUDIO_U8) &&
			(ioctl(audio_fd, AUDIO_SETINFO, &play_info) == 0) ) {
		if ( (play_info.play.sample_rate == spec->freq) &&
		     (play_info.play.precision == (spec->format&0xFF)) &&
		     (play_info.play.encoding == AUDIO_ENCODING_LINEAR) ) {
			ulaw_only = 0;
		} else {
			ioctl(audio_fd, AUDIO_SETINFO, &ulaw_info);
		}
	}
#endif /* AUDIO_GETINFO */

	/* We can actually convert on-the-fly to U-Law */
	if ( ulaw_only ) {
		ulaw_size = (spec->samples*1000)/(spec->freq/8);
		ulaw_buf = (Uint8 *)malloc(ulaw_size);
		if ( ulaw_buf == NULL ) {
			SDL_OutOfMemory();
			return(-1);
		}
		spec->format &= ~AUDIO_STEREO;
	}
#ifdef AUDIO_DEBUG
	fprintf(stderr, "Audio device %s U-Law only\n", 
				ulaw_only ? "is" : "is not");
#endif

	/* Update the fragment size as size in bytes */
	SDL_CalculateAudioSpec(spec);

	/* Allocate mixing buffer */
	mixlen = spec->size;
	mixbuf = (Uint8 *)SDL_AllocAudioMem(mixlen);
	if ( mixbuf == NULL ) {
		SDL_OutOfMemory();
		return(-1);
	}
	memset(mixbuf, spec->silence, spec->size);

	/* We're ready to rock and roll. :-) */
	return(0);
}

/************************************************************************/
/* This function (snd2au()) copyrighted:                                */
/************************************************************************/
/*      Copyright 1989 by Rich Gopstein and Harris Corporation          */
/*                                                                      */
/*      Permission to use, copy, modify, and distribute this software   */
/*      and its documentation for any purpose and without fee is        */
/*      hereby granted, provided that the above copyright notice        */
/*      appears in all copies and that both that copyright notice and   */
/*      this permission notice appear in supporting documentation, and  */
/*      that the name of Rich Gopstein and Harris Corporation not be    */
/*      used in advertising or publicity pertaining to distribution     */
/*      of the software without specific, written prior permission.     */
/*      Rich Gopstein and Harris Corporation make no representations    */
/*      about the suitability of this software for any purpose.  It     */
/*      provided "as is" without express or implied warranty.           */
/************************************************************************/

static Uint8 snd2au(int sample)
{

	int mask;

	if (sample < 0) {
		sample = -sample;
		mask = 0x7f;
	} else {
		mask = 0xff;
	}

	if (sample < 32) {
		sample = 0xF0 | 15 - (sample / 2);
	} else if (sample < 96) {
		sample = 0xE0 | 15 - (sample - 32) / 4;
	} else if (sample < 224) {
		sample = 0xD0 | 15 - (sample - 96) / 8;
	} else if (sample < 480) {
		sample = 0xC0 | 15 - (sample - 224) / 16;
	} else if (sample < 992) {
		sample = 0xB0 | 15 - (sample - 480) / 32;
	} else if (sample < 2016) {
		sample = 0xA0 | 15 - (sample - 992) / 64;
	} else if (sample < 4064) {
		sample = 0x90 | 15 - (sample - 2016) / 128;
	} else if (sample < 8160) {
		sample = 0x80 | 15 - (sample - 4064) /  256;
	} else {
		sample = 0x80;
	}
	return (mask & sample);
}
