/*
    NETLIB:  An example cross-platform network library for use with SDL
    Copyright (C) 1997  Sam Lantinga

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#include "SDL_types.h"

/* The maximum number of network channels (should be a multiple of 8) */
#define MAX_CHANNELS	32

/* The types of network transport supported by NETLIB */
/* If a transport type is unreliable, it means that packets sent via this
   transport may never arrive or may arrive out-of-order
 */
typedef enum {
	PROTO_UDP,			/* UDP transport -- unreliable */
	PROTO_TCP			/* TCP transport -- reliable */
} transport;

/* The basic packet datatype */
typedef struct {
	Uint8 *data;			/* Pointer to data buffer */
	Uint32 dlen;			/* Length of data buffer */
	Uint32 snum;			/* Sequence number of the packet */
} packet;
 
/* Initialize the network API */
extern int  Net_Init(void);
extern void Net_Quit(void);

/* Open a channel of 'type' transport type.
   If 'remote' is NULL, this is assumed to be a locally bound server port.
   If 'remote' is not NULL, a connection to the designated remote server
   and port is attempted.  Note that for unreliable connections, there may
   be nothing listening at the other end when the channel is created.
   The channel number (0-31) is returned if successful, or -1 on error.
*/
extern int Net_Open(transport type, char *remote, int port);

/* Accept a connection on the given network channel.
   This only works on channels that have been opened as server channels
   If 'remote' is not NULL,
	remote->dlen should be set to the length of the remote->data buffer
	remote->data contains the remote name
	remote->snum contains the remote port
   This returns a new channel for the connection, or -1 on error.
*/
extern int Net_Accept(int channel, packet *remote);

/* Send a packet over the given channel */
extern int Net_Send(int channel, packet *data);

/* Wait for a packet on the given channel
   The data packet should have dlen set to the size of the data buffer.
   Any packet that exceeds data->dlen in length will be truncated.
 */
extern int Net_Recv(int channel, packet *data);

/* Reset the sequence number of the channel to 0 */
extern void Net_Reset(int channel);

/* Wait for packets available on the channel mask designated by 'which'
   If timeout is not 0, Net_Wait() will return 0 if no packets have been
   received on the designated channels after 'timeout' milliseconds.
   Works just like the UNIX select() call:

   NetCMASK which;

   Net_ZERO(&which);
   Net_SET(&which, channel);
   Net_Wait(&which, 30);	// Wait for packets on channel for 30 ms
   if ( Net_ISSET(&which, channel) ) {
	Net_Recv(channel, packet);
   }
*/
#define Net_ZERO(mask)		memset(mask->bits, 0, MAX_CHANNELS/8)
#define Net_SET(mask, c)	mask->bits[c/8] |= (0x80>>(c%8))
#define NET_ISSET(mask, c)	((mask->bits[c/8]<<(c%8)) & 0x80)
typedef struct {
	Uint8 bits[MAX_CHANNELS/8];
} NetCMASK;
extern int Net_Wait(NetCMASK which, Uint32 timeout);

/* Close a network channel */
extern int Net_Close();

/* Get information about the remote end of the channel:
	remote->dlen should be set to the length of the remote->data buffer
	remote->data contains the remote name
	remote->snum contains the remote port
*/
extern int Net_Info(int channel, packet *remote);

/* Write a 16/32 bit value to network packet buffer */
void Net_Write16(Uint16 value, Uint8 *area);
void Net_Write32(Uint32 value, Uint8 *area);

/* Read a 16/32 bit value from network packet buffer */
Uint16 Net_Read16(Uint8 *area);
Uint32 Net_Read32(Uint8 *area);

