
/* A simple program to test the text rendering feature of the TTF library */

#include <stdio.h>

#include "SDL.h"
#include "SDLttf.h"

#define DEFAULT_PTSIZE	18
#define DEFAULT_TEXT	"The quick brown fox jumped over the lazy dog"

static char *Usage = "Usage: %s <font>.ttf [ptsize] [text]\n";

int main(int argc, char *argv[])
{
	SDL_Surface *screen;
	TTF_Font *font;
	SDL_Surface *text, *temp;
	int ptsize;
	int i, done;
	int rdiff, gdiff, bdiff;
	SDL_Color colors[5];
	SDL_Color white = { 0xFF, 0xFF, 0xFF, 0 };
	SDL_Color black = { 0x00, 0x00, 0x00, 0 };
	Uint32 white_pixel;
	SDL_Rect dstrect;
	SDL_Event event;
	char *message, string[128];

	/* Check usage */
	if ( argc == 1 ) {
		fprintf(stderr, Usage, argv[0]);
		exit(1);
	}

	/* Initialize SDL */
	if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr, "Couldn't initialize SDL: %s\n",SDL_GetError());
		exit(2);
	}
	atexit(SDL_Quit);

	/* Initialize the TTF library */
	if ( TTF_Init() < 0 ) {
		fprintf(stderr, "Couldn't initialize TTF: %s\n",SDL_GetError());
		exit(2);
	}
	atexit(TTF_Quit);

	/* Set a 640x480x8 video mode */
	screen = SDL_SetVideoMode(640, 480, 8, SDL_SWSURFACE);
	if ( screen == NULL ) {
		fprintf(stderr, "Couldn't set 640x480x8 video mode: %s\n",
							SDL_GetError());
		exit(2);
	}

	/* Set a palette that is good for black text */
	rdiff = white.r - black.r;
	gdiff = white.g - black.g;
	bdiff = white.b - black.b;
	for ( i=0; i<5; ++i ) {
		colors[i].r = black.r + (i*rdiff)/4;
		colors[i].g = black.g + (i*gdiff)/4;
		colors[i].b = black.b + (i*bdiff)/4;
	}
	SDL_SetColors(screen, colors, 10, 5);

	/* Clear the background to white */
	white_pixel = SDL_MapRGB(screen->format, white.r, white.g, white.b);
	if ( SDL_LockSurface(screen) == 0 ) {
		Uint16 h;
		Uint8 *buffer;

		buffer = (Uint8 *)screen->pixels;
		h = screen->h;
		while ( h-- ) {
			memset(buffer, white_pixel, screen->w);
			buffer += screen->pitch;
		}
		SDL_UnlockSurface(screen);
	}
		
	/* Open the font file with the requested point size */
	ptsize = 0;
	if ( argv[2] ) {
		ptsize = atoi(argv[2]);
	}
	if ( ptsize == 0 ) {
		i = 2;
		ptsize = DEFAULT_PTSIZE;
	} else {
		i = 3;
	}
	font = TTF_OpenFont(argv[1], ptsize);
	if ( font == NULL ) {
		fprintf(stderr, "Couldn't load %d pt font from %s: %s\n",
					ptsize, argv[1], SDL_GetError());
		exit(2);
	}

	/* Show which font file we're looking at */
	sprintf(string, "Font file: %s", argv[1]);  /* possible overflow */
	text = TTF_RenderText(font, string, black, white);
	if ( text != NULL ) {
		if ( SDL_MapSurface(text, screen->format) == 0 ) {
			dstrect.x = 4;
			dstrect.y = 4;
			dstrect.w = text->w;
			dstrect.h = text->h;
			SDL_BlitSurface(text, NULL, screen, &dstrect);
		}
		SDL_FreeSurface(text);
	}
	
	/* Render and center the message */
	if ( argv[i] ) {
		message = argv[i];
	} else {
		message = DEFAULT_TEXT;
	}
	text = TTF_RenderText(font, message, black, white);
	if ( text == NULL ) {
		fprintf(stderr, "Couldn't render text: %s\n", SDL_GetError());
		TTF_CloseFont(font);
		exit(2);
	}
	dstrect.x = (screen->w - text->w)/2;
	dstrect.y = (screen->h - text->h)/2;
	dstrect.w = text->w;
	dstrect.h = text->h;
	printf("Font is generally %d big, and string is %hd big\n",
						TTF_FontHeight(font), text->h);

	/* Map and blit the text surface */
	if ( SDL_MapSurface(text, screen->format) < 0 ) {
		fprintf(stderr, "Couldn't map text to display: %s\n", 
								SDL_GetError());
		TTF_CloseFont(font);
		exit(2);
	}
	if ( SDL_BlitSurface(text, NULL, screen, &dstrect) < 0 ) {
		fprintf(stderr, "Couldn't blit text to display: %s\n", 
								SDL_GetError());
		TTF_CloseFont(font);
		exit(2);
	}
	SDL_UpdateRect(screen, 0, 0, 0, 0);

	/* Set the text colorkey and convert to display format */
	if ( SDL_SetColorKey(text, SDL_SRCCOLORKEY|SDL_RLEACCEL, 0) < 0 ) {
		fprintf(stderr, "Warning: Couldn't set text colorkey: %s\n",
								SDL_GetError());
	}
	temp = SDL_DisplayFormat(text);
	if ( temp != NULL ) {
		SDL_FreeSurface(text);
		text = temp;
	}

	/* Wait for a keystroke, and blit text on mouse press */
	done = 0;
	while ( ! done ) {
		if ( SDL_WaitEvent(&event) < 0 ) {
			fprintf(stderr, "SDL_PullEvent() error: %s\n",
								SDL_GetError());
			done = 1;
			continue;
		}
		switch (event.type) {
			case SDL_MOUSEBUTTONDOWN:
				dstrect.x = event.button.x - text->w/2;
				dstrect.y = event.button.y - text->h/2;
				dstrect.w = text->w;
				dstrect.h = text->h;
				if ( SDL_BlitSurface(text, NULL, screen,
							&dstrect) == 0 ) {
					SDL_UpdateRects(screen, 1, &dstrect);
				} else {
					fprintf(stderr,
					"Couldn't blit text to display: %s\n", 
								SDL_GetError());
				}
				break;
				
			case SDL_KEYDOWN:
			case SDL_QUIT:
				done = 1;
				break;
			default:
				break;
		}
	}
	SDL_FreeSurface(text);
	TTF_CloseFont(font);
	exit(0);
}
